from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate

logger = logging.getLogger(__name__)


class Logs(Command):
    def getShellOptions(self, optparser):
        optparser.add_option("-n", "--lines", dest="lines",
                             help="Output the last N lines", default=None, metavar="N")
        optparser.add_option("-p", "--pattern", dest="pattern",
                             help="Use a log pattern instead of arguments", default=None)
        optparser.add_option("-l", "--list", dest="list",
                             help="List available environment logs", default=None, action="store_true")
        return optparser

    def getUsage(self):
        return "substance logs [options] [container] [logname]"

    def getHelpTitle(self):
        return "Display current subenv logs"

    def main(self):

        op = self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.envLoadCurrent)

        if self.getOption('list'):
            op = op.bind(Engine.envListLogs, parts=self.getArgs(),
                         pattern=self.getOption('pattern'))
        else:
            op = op.bind(Engine.envLogs, parts=self.getArgs(), pattern=self.getOption(
                'pattern'), lines=self.getOption('lines'))

        op = op.catch(self.exitError)
        return op



from substance import (Engine, Command)
from substance.exceptions import (LinkCommandError)


class Run(Command):

    def getUsage(self):
        return "substance run JOBCONTAINER [ARGS...]"

    def getHelpTitle(self):
        return "Run a job container with arguments"

    def getShellOptions(self, optparser):
        return optparser

    # Disable parsing for this command
    def execute(self, args=None):
        self.input = args
        self.args = args

        self.main()

    def main(self):

        if len(self.args) == 0:
            return self.exitError("Please provide a jobcontainer name to run.")
        
        task = self.args[0]
        args = self.args[1:]

        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.envRun, task=task, args=args) \
            .catch(self.onEngineException)

    def onEngineException(self, err):
        return self.exitError(err)



from substance import (Command, Engine)
from substance.logs import (logging)

logger = logging.getLogger(__name__)


class Expose(Command):

    def getUsage(self):
        return "substance expose [options] LOCALPORT[:PUBLICPORT]"

    def getHelpTitle(self):
        return "Expose an engine port as a port on the host machine."

    def getShellOptions(self, optparser):
        optparser.add_option(
            "-s",
            "--scheme",
            dest="scheme",
            help="Specify the uri scheme (e.g. wss) that this port will be used for.",
            default=None)
        return optparser

    def main(self):
        portdef = self.getArg(0)
        if not portdef:
            return self.exitHelp("You must provide a port description!")

        if ":" in portdef:
            (loc, pub) = portdef.split(":", 2)
            local_port = int(loc)
            public_port = int(pub)
        else:
            local_port = int(portdef)
            public_port = int(portdef)

        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.exposePort, local_port, public_port, self.getScheme(public_port))

    def getScheme(self, public_port):
        scheme = self.getOption('scheme')
        if scheme:
            return scheme
        # Try a guess!
        return {
            20: "ftp",
            21: "ftp",
            22: "ssh",
            23: "telnet",
            25: "smtp",
            53: "dns",
            70: "gopher",  # Yeah that's right. Long live Gopher!
            79: "finger",
            80: "http",
            123: "ntp",
            143: "imap",
            161: "snmp",
            162: "snmp",
            194: "irc",
            199: "snmp",
            220: "imap",
            443: "https",
            445: "smb",
            465: "smtp",
            873: "rsync",
            944: "nfs",
            992: "telnet",
            994: "ircs",
            8080: "http",
            6660: "irc",
            6661: "irc",
            6662: "irc",
            6663: "irc",
            6664: "irc",
            6665: "irc",
            6666: "irc",
            6667: "irc",
            6668: "irc",
            6669: "irc"
        }.get(public_port, "ws")



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate


class Start(Command):
    def getShellOptions(self, optparser):
        optparser.add_option("-r", "--reset", dest="reset",
                             help="Stop & remove all containers before starting", default=False, action="store_true")
        return optparser

    def getUsage(self):
        return "substance start [options] [CONTAINER...]"

    def getHelpTitle(self):
        return "Start all or specific container(s)"

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.envStart, reset=self.getOption('reset'), containers=self.args) \
            .catch(self.exitError)



import os
import sys
import platform
import time
import logging
import shutil
import shlex
import subprocess
from subprocess import call, Popen, check_output, CalledProcessError
from substance.exceptions import (
    FileSystemError, ShellCommandError, UserInterruptError)
from substance.monads import *
from threading import Thread
from substance.platform import (isCygwin)

# pylint: disable=W0232

logger = logging.getLogger(__name__)


class Shell(object):

    @staticmethod
    def printConfirm(msg, assumeYes=False):
        if assumeYes:
            return OK(True)

        logger.info(msg)
        try:
            res = input('Proceed? [N/y] ')
            if res.lower().startswith('y'):
                logger.info('... proceeding')
                return OK(True)
            return Fail(UserInterruptError(message="User interrupted."))
        except KeyboardInterrupt as err:
            return Fail(UserInterruptError(message="User interrupted."))

    @staticmethod
    def call(cmd, cwd=None, shell=True, sudo=False):
        try:
            mustSleep = False
            if sudo and not shell:
                if isCygwin():
                    mustSleep = True
                    cmd = ["cygstart", "--action=runas"] + cmd
                else:
                    cmd = ["sudo"] + cmd
            logger.debug("COMMAND[%s]: %s", cwd, cmd)
            returncode = call(cmd, shell=shell, cwd=cwd)
            if returncode == 0:
                # Allow command to finish on Windows (cygstart exits early)
                return OK(None).then(lambda: None if not mustSleep else time.sleep(1))
            else:
                return Fail(ShellCommandError(code=returncode))
        except CalledProcessError as err:
            return Fail(ShellCommandError(code=err.returncode, message=err.output, stdout=err.output))

    @staticmethod
    def command(cmd):
        logger.debug("COMMAND: %s", cmd)
        try:
            out = check_output(cmd, shell=True)
            return OK(out.decode().strip())
        except CalledProcessError as err:
            return Fail(ShellCommandError(code=err.returncode, message=err.output, stdout=err.output))

    @staticmethod
    def procCommand(cmd, cwd=None):
        try:
            logger.debug("PROC COMMAND: %s", cmd)
            proc = Popen(shlex.split(cmd), shell=False,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
            pout, perr = proc.communicate()
            cmdOut = pout.decode()
            cmdErr = perr.decode()
            if proc.returncode == 0:
                return OK({"stdout": cmdOut, "stderr": cmdErr})
            else:
                return Fail(ShellCommandError(code=proc.returncode, message=cmdOut, stdout=cmdOut, stderr=cmdErr))
        except KeyboardInterrupt:
            logger.info("CTRL-C Received...Exiting.")
            return Fail(UserInterruptError())

    @staticmethod
    def readAppendFlush(stream, buffer=b'', amount=1):
        the_bytes = stream.read(amount)
        if the_bytes:
            the_bytes += byte
            sys.stdout.write(the_bytes.encode())
            sys.stdout.flush()
        return buffer

    @staticmethod
    def streamCommand(cmd, cwd=None, shell=False):
        try:
            logger.debug("STREAM COMMAND: %s", cmd)
            proc = Popen(shlex.split(cmd), shell=shell,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
            stdout = b''
            stderr = b''
            while True:
                stdout = readAppendFlush(sys.stdout, stdout, 1)
                stderr = readAppendFlush(sys.stderr, stderr, 1)

                if not d  and not de and proc.poll() is not None:
                    break

            stdout = stdout.encode()
            stderr = stderr.encode()

            if proc.returncode == 0:
                return OK({"stdout": stdout, "stderr": stderr})
            else:
                return Fail(ShellCommandError(code=proc.returncode, message=stdout, stdout=stdout, stderr=stderr))
        except OSError as err:
            err.strerror = "Error running '%s': %s" % (cmd, err.strerror)
            return Fail(err)
        except KeyboardInterrupt:
            logger.info("CTRL-C Received...Exiting.")
            return Fail(UserInterruptError())

    @staticmethod
    def execvp(cmdPath, cmdArgs, cmdEnv=None, sudo=False):
        logger.debug("EXECVPE: `%s` with environment %s",
                     subprocess.list2cmdline(cmdArgs), cmdEnv)
        if sudo:
            cmdPath = "sudo"
            cmdArgs = ["sudo"] + cmdArgs
        if cmdEnv:
            os.execvpe(cmdPath, cmdArgs, cmdEnv)
        else:
            os.execvp(cmdPath, cmdArgs)

    @staticmethod
    def stringify(cmd, env={}):
        envStr = ""
        if env:
            envStr = " ".join(["%s=%s" % (k, subprocess.list2cmdline([env[k]])) for k in env]) + " "
        return "%s%s" % (envStr, subprocess.list2cmdline(cmd))

    @staticmethod
    def chmod(path, mode):
        try:
            os.chmod(path, mode)
            return OK(path)
        except Exception as err:
            return Fail(ShellCommandError(code=1, message="Failed to chmod %s: %s" % (path, err)))

    @staticmethod
    def makeDirectory(path, mode=0o750):
        if not os.path.exists(path):
            try:
                os.makedirs(path, mode)
                return OK(path)
            except Exception as err:
                return Fail(ShellCommandError(code=1, message="Failed to create %s: %s" % (path, err)))
        return OK(path)

    @staticmethod
    def copyFile(src, dst):
        return Try.attempt(lambda: shutil.copy(src, dst))

    @staticmethod
    def pathExists(path):
        return OK(path) if os.path.exists(path) else Fail(FileSystemError("Path %s does not exist." % path))

    @staticmethod
    def rmFile(path):
        if os.path.isdir(path):
            return Fail(FileSystemError("%s is a directory.") % path)
        return OK(os.remove(path))

    @staticmethod
    def nukeDirectory(path):
        try:
            if path and path is not '/' and path is not 'C:\\':
                shutil.rmtree(path)
            return OK(None)
        except Exception as err:
            return Fail(ShellCommandError(code=1, message="Failed to rmtree: %s: %s" % (path, err)))



import unittest
import os
import tempfile
import tests
import re
from substance.core import Core
from substance.engine import (EngineProfile, Engine)
from substance.shell import Shell
from substance.monads import *
from substance.constants import *


class TestEngine(tests.TestBase):

    core = None
    engine = None
    basePath = None
    projectsPath = None

    vmTest = True

    @classmethod
    def setUpClass(cls):
        cls.basePath = cls.addTemporaryDir()
        cls.projectsPath = cls.addTemporaryDir()

        cls.core = Core(basePath=cls.basePath)
        cls.core.initialize().catch(cls.raiser)

        cls.engine = cls.core.createEngine("testEngine", config={"driver": "virtualbox", "projectsPath": cls.projectsPath})  \
            .bind(Engine.loadConfigFile)  \
            .catch(TestEngine.raiser) \
            .getOK()

    @classmethod
    def tearDownClass(cls):
        if cls.engine and cls.engine.isProvisioned():
            cls.engine.deprovision()

        if cls.basePath:
            Shell.nukeDirectory(cls.basePath).catch(cls.raiser)
        if cls.projectsPath:
            Shell.nukeDirectory(cls.projectsPath).catch(cls.raiser)

    def testInitSequence(self):
        if not self.vmTest:
            return
        self.doProvision()
        self.doStart()
        self.doSuspend()
        self.doRestart()
        self.doDeprovision()

    def testLaunch(self):
        if not self.vmTest:
            return
        self.doLaunch()

    def doProvision(self):
        op = self.engine.provision()
        self.assertIsInstance(op, OK)
        self.assertTrue(self.engine.isProvisioned())
        self.assertIsNotNone(self.engine.getDriverID())
        self.checkStateMatch(EngineStates.STOPPED)

    def doStart(self):
        op = self.engine.start()
        self.assertIsInstance(op, OK)
        self.checkStateMatch(EngineStates.RUNNING)

    def doSuspend(self):
        op = self.engine.suspend()
        self.assertIsInstance(op, OK)
        self.checkStateMatch(EngineStates.SUSPENDED)

    def doRestart(self):
        op = self.engine.start()
        self.assertIsInstance(op, OK)
        self.checkStateMatch(EngineStates.RUNNING)

    def doDeprovision(self):
        op = self.engine.deprovision()
        self.assertIsInstance(op, OK)
        self.assertFalse(self.engine.isProvisioned())
        self.assertIsNone(self.engine.getDriverID())
        self.checkStateMatch(EngineStates.INEXISTENT)

    def doLaunch(self):
        op = self.engine.launch()
        self.assertIsInstance(op, OK)

    def checkStateMatch(self, stateMatch):
        state = self.engine.fetchState()
        self.assertIsInstance(state, OK)
        self.assertEqual(state.getOK(), stateMatch)





import sys
import os
import re
import logging
import pkgutil
import importlib
import traceback
from collections import OrderedDict
from importlib import import_module
import optparse
from getpass import getpass

from substance import Shell
from substance.exceptions import InvalidCommandError

logger = logging.getLogger('substance')


class Parser(optparse.OptionParser):
    pass
#  def format_description(self, formatter):
#    return self.description if self.description else ""

    def format_epilog(self, formatter):
        return "\n" + self.epilog if self.epilog else ""


class PassThroughParser(Parser):
    def _process_long_opt(self, rargs, values):
        try:
            optparse.OptionParser._process_long_opt(self, rargs, values)
        except optparse.BadOptionError as err:
            self.largs.append(err.opt_str)

    def _process_short_opts(self, rargs, values):
        try:
            optparse.OptionParser._process_short_opts(self, rargs, values)
        except optparse.BadOptionError as err:
            self.largs.append(err.opt_str)


class CLI(object):
    def __init__(self):
        self.args = None
        self.options = None
        self.input = None
        self.parser = None
        self.name = None
        self.parent = None

    def getHelp(self):
        """Returns the help description for this particular command"""
        return self.getParser().format_help()

    def getHelpTitle(self):
        return ""

    def getHelpDetails(self):
        return ""

    def getUsage(self):
        return "command: USAGE [options]"

    def getParserClass(self):
        return Parser

    def getParser(self, interspersed=True):
        usage = self.getUsage()
        parserClass = self.getParserClass()
        parser = parserClass(usage=usage, conflict_handler="resolve", add_help_option=False,
                             description=self.getHelpTitle(), epilog=self.getHelpDetails())
        self.getShellOptions(parser)
        if interspersed:
            parser.enable_interspersed_args()
        else:
            parser.disable_interspersed_args()
        return parser

    def parseShellInput(self, interspersed=True):
        """Return the options and arguments for this command as a tuple"""
        usage = self.getUsage()
        parser = self.getParser(interspersed)
        (opts, args) = parser.parse_args(self.input)
        return (opts, args)

    def getShellOptions(self, optparser):
        return optparser

    def execute(self, args=None):
        self.input = args
        (self.options, self.args) = self.parseShellInput()
        return self.main()

    def initialize(self):
        pass

    def main(self):
        """The main function body"""
        pass

    def getOptions(self):
        return self.options

    def hasOption(self, key):
        if self.options and hasattr(self.options, key):
            return True

    def getOption(self, key):
        if self.hasOption(key):
            return getattr(self.options, key, None)
        return None

    def getArgs(self):
        return self.args

    def hasArg(self, argno):
        if len(self.args) < argno + 1:
            return False
        return True

    def getArg(self, argno):
        if self.hasArg(argno):
            return self.args[argno]
        return None

    def read(self, msg=""):
        return input(msg)

    def readln(self, msg=""):
        return input(msg + "\n")

    def readpwd(self, msg=""):
        return getpass(msg)

    def readpwdln(self, msg=""):
        return getpass(msg + "\n")

    def exitOK(self, msg=None):
        if msg:
            logger.info(msg)
        sys.exit(0)

    def exitHelp(self, msg=None, code=1):
        logger.info(self.getHelp())
        logger.info("")
        if msg:
            logger.info(msg)
        sys.exit(1)

    def exitError(self, msg=None, code=1):
        if msg and isinstance(msg, Exception):
            logger.error("%s (%s)" % (msg, type(msg).__name__))
        elif msg:
            logger.error(msg)
        sys.exit(1)


class Program(CLI):
    def __init__(self):
        super(Program, self).__init__()
        self.cmdString = None
        self.commandInput = None
        self.command = None
        self.core = None
        self.commands = OrderedDict()
        self.addCommand('help', 'substance.command.help')

    def setupCommands(self):
        pass

    def getCommands(self):
        k = list(self.commands.keys())
        v = [self.getCommand(x) for x in k]
        return OrderedDict(list(zip(k, v)))

    def getCommand(self, commandName):
        if commandName in self.commands:
            commandModule = self.commands[commandName]
            commandClass = commandName.title()
        else:
            commandModule = 'substance.command.fallback'
            commandClass = 'Fallback'

        module_ = importlib.import_module(commandModule)
        class_ = getattr(module_, commandClass)
        command = class_()
        command.parent = self
        command.name = commandName
        return command

    def addCommand(self, commandName, commandModule):
        if commandName in self.commands:
            raise ValueError(
                "Command '%s' is already registered." % commandName)
        self.commands[commandName] = commandModule

    def removeCommand(self, commandName):
        if commandName not in self.commands:
            raise ValueError("Command '%s' is not registered." % commandName)
        self.commands.pop(commandName, None)

    def getHelpDetails(self):
        helpUsage = "Commands:\n\n"
        for name, command in self.getCommands().items():
            helpUsage += "  %-20s%s\n" % (name, command.getHelpTitle())
        return helpUsage

    def setupEnv(self):
        """Setup the Environment """

    def setupLogging(self):
        if self.getOption('debug'):
            logging.getLogger('substance').handlers[0].setLevel(logging.DEBUG)
            logging.getLogger('subwatch').setLevel(logging.DEBUG)
            logging.getLogger('watchdog').setLevel(logging.INFO)
            logging.getLogger('paramiko').setLevel(logging.INFO)

    def execute(self, args=None):
        self.input = args
        (self.options, self.args) = self.parseShellInput(False)
        self.main()

    def initialize(self):
        self.setupLogging()
        self.setupEnv()
        self.setupCommands()

    def main(self):
        try:
            self.initialize()

            if not len(self.args) > 0:
                self.cmdString = 'help'
                self.commandInput = []
            else:
                self.cmdString = self.args[0]
                self.commandInput = self.args[1:]

            self.runCommand(self.cmdString)
        except Exception as err:
            logger.error(traceback.format_exc())
            self.exitError("Error running command %s: %s" %
                           (self.cmdString, err), code=2)

    def runCommand(self, commandName):
        command = self.getCommand(commandName)
        self.command = self.initCommand(command)
        self.command.execute(self.commandInput)


class Command(CLI):

    def __init__(self):
        super(Command, self).__init__()

    def ask(self, msg):
        return Shell.printConfirm(msg, assumeYes=self.core.getAssumeYes())

    # ---- Validation Helpers

    def getInputName(self):
        if len(self.args) <= 0:
            return self.exitError("Please specify an engine name.")

        name = self.args[0]
        if not name or not self.validateEngineName(name):
            return self.exitError("Invalid name specified. Please use only letters, numbers, dash or underscores.")
        return name

    def validateEngineName(self, name):
        return bool(re.match(r'^[a-zA-Z0-9\-_]*$', name))

    def validateDriver(self, driver):
        return self.core.validateDriver(driver)

    def validateInteger(self, value):
        try:
            val = int(value)
            return True
        except ValueError:
            return False


class SubProgram(Program):
    def initialize(self):
        self.setupCommands()



import os
from substance.logs import *
from substance import (Command, Box, EngineProfile)
from substance.exceptions import (InvalidOptionError)


class Create(Command):

    def getUsage(self):
        return "substance engine init [options] [ENGINE NAME]"

    def getHelpTitle(self):
        return "Create a new engine configuration"

    def getShellOptions(self, optparser):
        optparser.set_description("Create a new substance engine")
        optparser.add_option("--devroot", dest="devroot",
                             help="Path to local devroot directory.")
        optparser.add_option('--devroot-mode', dest="devroot_mode",
                             help="devroot sync mode", default="unison")
        optparser.add_option("--mount", dest="mounts",
                             help="Mount host path to engine path", nargs=10)
        optparser.add_option("--driver", dest="driver",
                             help="Virtualization driver for this engine")
        optparser.add_option("--memory", type="int",
                             dest="memory", help="Machine memory allocation")
        optparser.add_option("--cpus", type="int",
                             dest="cpus", help="Machine vCPU allocation")
        optparser.add_option("--box", type="str",
                             dest="box", help="Engine box image")
        return optparser

    def main(self):

        name = self.getInputName()

        options = self.buildConfigFromArgs().bind(self.buildProfileFromArgs) \
            .catch(self.exitError).getOK()

        self.core.createEngine(name, config=options['config'], profile=options['profile']) \
            .bind(dinfo("Engine \"%s\" has been created", name)) \
            .catch(self.exitError) \


    def buildConfigFromArgs(self, config={}):
        opts = {}

        boxName = self.core.getDefaultBoxString(
        ) if not self.options.box else self.options.box
        box = self.core.readBox(boxName)
        if box.isFail():
            return box
        opts['box'] = box.getOK().boxstring

        if self.options.driver:
            if not self.validateDriver(self.options.driver):
                return Fail(InvalidOptionError("Driver %s is not valid." % self.options.driver))
            opts['driver'] = self.options.driver

        if self.options.devroot:
            if not os.path.isdir(self.options.devroot):
                return Fail(InvalidOptionError("Devroot path %s does not exist." % self.options.devroot))
            opts['devroot'] = {} if 'devroot' not in opts else opts['devroot']
            opts['devroot']['path'] = self.options.devroot

        if self.options.devroot_mode:
            # XXX Fix hardcoded values.
            if self.options.devroot_mode not in ['rsync', 'sharedfolder', 'unison']:
                return Fail(InvalidOptionError("Devroot mode '%s' is not valid."))
            opts['devroot'] = {} if 'devroot' not in opts else opts['devroot']
            opts['devroot']['mode'] = self.options.devroot_mode

        config['config'] = opts
        return OK(config)

    def buildProfileFromArgs(self, config={}):
        profile = EngineProfile()
        if self.options.memory and self.validateInteger(self.options.memory):
            profile.memory = self.options.memory
        if self.options.cpus and self.validateInteger(self.options.cpus):
            profile.cpus = self.options.cpus

        config['profile'] = profile
        return OK(config)



from substance.monads import *
from substance.logs import *
from substance import (Command, Shell, Engine)


class Halt(Command):

    def getUsage(self):
        return "substance engine halt [options] [ENGINE NAME]"

    def getHelpTitle(self):
        return "Halt an engine's vm gracefully or forcefully."

    def getShellOptions(self, optparser):
        optparser.add_option("-f", "--force", dest="force",
                             help="Force power off", action="store_true")
        return optparser

    def main(self):

        name = self.getInputName()

        forced = self.options.force
        if forced and not Shell.printConfirm("You are about to force halt engine \"%s\"." % name):
            self.exitOK("User cancelled.")

        self.core.loadEngine(name) \
            .bind(Engine.loadConfigFile) \
            .bind(defer(Engine.stop, force=forced)) \
            .catch(self.exitError)




import sys
import logging
from collections import OrderedDict
from substance import (Program, Core)
from substance.utils import getPackageVersion
import substance.command
import substance.logs


class SubstanceCLI(Program):
    """Substance CLI command"""

    def __init__(self):
        super(SubstanceCLI, self).__init__()
        self.name = 'substance'

    def setupCommands(self):
        self.addCommand('use', 'substance.command.use')
        self.addCommand('launch', 'substance.command.launch')
        self.addCommand('suspend', 'substance.command.suspend')
        self.addCommand('halt', 'substance.command.halt')
        self.addCommand('switch', 'substance.command.switch')
        self.addCommand('start', 'substance.command.start')
        self.addCommand('stop', 'substance.command.stop')
        self.addCommand('restart', 'substance.command.restart')
        self.addCommand('recreate', 'substance.command.recreate')
        self.addCommand('ssh', 'substance.command.ssh')
        self.addCommand('exec', 'substance.command.exec')
        self.addCommand('run', 'substance.command.run')
        self.addCommand('status', 'substance.command.status')
        self.addCommand('logs', 'substance.command.logs')
        self.addCommand('cleanup', 'substance.command.cleanup')
        self.addCommand('sync', 'substance.command.sync')
        self.addCommand('engine', 'substance.command.engine')
        self.addCommand('box', 'substance.command.box')
        self.addCommand('aliases', 'substance.command.aliases')
        self.addCommand('hatch', 'substance.command.hatch')
        self.addCommand('expose', 'substance.command.expose')
        return self

    def getShellOptions(self, optparser):
        optparser.add_option("-e", "--engine", dest="engine",
                             help="Engine to run this command on", default=None)
        optparser.add_option("-f", dest="configFile",
                             help="Override default config file")
        optparser.add_option(
            "-d", dest="debug", help="Activate debugging output", default=False, action="store_true")
        optparser.add_option(
            "-y", dest="assumeYes", help="Assume yes when prompted", default=False, action="store_true")
        return optparser

    def getUsage(self):
        return "substance [options] COMMAND [command-options]"

    def getHelpTitle(self):
        version = getPackageVersion()
        return "Local docker-based development environments (version: %s)" % version

    def initCommand(self, command):
        core = Core()
        if self.getOption('assumeYes'):
            core.setAssumeYes(True)

        core.initialize().catch(self.exitError)

        command.core = core
        return command


def cli():
    args = sys.argv[1:]
    prog = SubstanceCLI()
    prog.execute(args)



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate


class Use(Command):
    def getShellOptions(self, optparser):
        optparser.add_option("-e", "--env", dest="env",
                             help="Also switch to this environment", default=False)
        return optparser

    def getUsage(self):
        return "substance use [options] ENGINE-NAME"

    def getHelpTitle(self):
        return "Specify which engine is used by default"

    def main(self):
        name = self.getInputName()
        return self.core.loadEngine(name) \
            .bind(Engine.loadConfigFile) \
            .bind(self.core.setUse, subenvName=self.getOption('env')) \
            .then(self.core.config.saveConfig) \
            .catch(self.exitError)



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate
from substance.command.command import PassThroughParser


class Docker(Command):

    def getParserClass(self):
        return PassThroughParser

    def getShellOptions(self, optparser):
        return optparser

    def getUsage(self):
        return "substance docker [options] COMMAND"

    def getHelpTitle(self):
        return "Run a docker command on an engine"

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.envLoadCurrent) \
            .bind(Engine.envDocker, command=' '.join(self.args)) \
            .catch(self.exitError)



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate

logger = logging.getLogger(__name__)


class Ls(Command):
    def getUsage(self):
        return "substance engine ls [options]"

    def getHelpTitle(self):
        return "List substance engines"

    def main(self):
        self.core.getEngines() \
            .mapM(self.core.loadEngine) \
            .mapM(Engine.loadConfigFile) \
            .mapM(Engine.loadState) \
            .mapM(self.tabulateEngine) \
            .bind(self.tabulateEngines) \
            .catch(self.exitError) \
            .bind(logger.info)

    def tabulateEngine(self, engine):
        state = engine.state if engine.state else "-"
        prov = "yes" if engine.isProvisioned() else "no"
        ip = engine.getPublicIP() if engine.getPublicIP() else "-"
        return OK([engine.getName(), prov, state, ip, "-"])

    def tabulateEngines(self, engines):
        headers = ["NAME", "PROVISIONED", "STATE", "IP", "ERRORS"]
        table = tabulate(engines, headers=headers, tablefmt="plain")
        return OK(table)



import os
import logging
import platform

from substance.utils import getSupportFile
from substance.constants import Syncher
from substance.shell import Shell
from substance.syncher import BaseSyncher
from substance.monads import Try
from substance.link import Link
from substance.path import (getHomeDirectory, inner, outer)
from substance.platform import (isWithinWindowsSubsystem)

logger = logging.getLogger(__name__)


class UnisonSyncher(BaseSyncher):

    def __init__(self, engine, keyfile):
        super(UnisonSyncher, self).__init__(engine, keyfile)
        self.ignoreArchives = False

    def start(self, direction, path=''):
        self.ensureKeyPerms()
        unisonPath = self.getUnisonBin()
        unisonArgs = self.getUnisonArgs(direction, path)
        unisonEnv = self.getUnisonEnv()
        if self.ignoreArchives:
            # It seems that in certain cases, unison does not observe the -ignorearchives flag correctly
            # So to make sure, we forcibly delete previous archives on both sides
            res = Try.sequence([
                # Delete host archives
                Shell.call(["rm", "-rf", unisonEnv['UNISON']], shell=False),
                # Delete guest archives
                self.engine.readLink().bind(Link.runCommand, 'rm -rf /substance/.unison')
            ])
            if res.isFail():
                return res
        logger.info("Syncing local directory %s to remote directory %s", unisonArgs[-2], unisonArgs[-1])
        logger.debug("EXEC: %s", Shell.stringify([unisonPath] + unisonArgs, unisonEnv))
        os.execve(unisonPath, unisonArgs, unisonEnv)

    def getUnisonBin(self):
        unisonDir = self.getUnisonSupportDirectory()
        logger.debug("Dir: %s", unisonDir)
        return os.path.join(unisonDir, 'unison')

    def getUnisonArgs(self, direction, path=''):
        folder = self.engine.getEngineFolders()[0]
        localRoot = inner(os.path.join(folder.hostPath, path))
        remoteRoot = 'ssh://substance@%s/%s/%s' % (
            self.engine.getSSHIP(), folder.guestPath.rstrip('/'), path.lstrip('/'))

        # Direction arguments
        if direction == Syncher.UP:
            directionArgs = ['-nocreation', localRoot,
                             '-nodeletion', localRoot, '-noupdate', localRoot]
        elif direction == Syncher.DOWN:
            directionArgs = ['-nocreation', remoteRoot,
                             '-nodeletion', remoteRoot, '-noupdate', remoteRoot]
        else:
            directionArgs = ['-prefer', 'newer', '-copyonconflict']

        optArgs = []
        if isWithinWindowsSubsystem():
            optArgs = ['-perms', '0']

        # User args
        userArgs = folder.syncArgs

        # Other arguments
        rootArgs = [localRoot, remoteRoot]
        ignoreArgs = [item for sublist in [['-ignore', 'Name ' + excl]
                                           for excl in folder.excludes] for item in sublist]

        # SSH config
        transport = "-p %s -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null" % self.engine.getSSHPort()
        if self.keyfile is not None:
            transport += " -i %s" % inner(self.keyfile)

        # Assemble everything
        args = ['-batch', '-repeat', 'watch', '-sshargs', transport] + \
            optArgs + userArgs + directionArgs + ignoreArgs + rootArgs
        if self.ignoreArchives and '-ignorearchives' not in args:
            args.insert(0, '-ignorearchives')
        return args

    def getUnisonEnv(self):
        # Add the unison-fsmonitor binary to PATH
        path = inner(os.path.join(self.getUnisonSupportDirectory(), os.environ.get('PATH', '')))
        # Tell unison to save all replica state to ~/.substance/unison
        unisonDir = inner(os.path.join(self.engine.core.getBasePath(), 'unison'))
        homeDir = getHomeDirectory()
        return {
            'UNISON': unisonDir, 
            'PATH': path, 
            'HOME': homeDir, 
            'SSH_AUTH_SOCK': os.environ.get('SSH_AUTH_SOCK', '')
        }

    def getUnisonSupportDirectory(self):
        osTarget = None
        osName = os.name
        psys = platform.system()
        if osName == 'posix':
            if "CYGWIN" in psys:
                osTarget = 'windows'
            elif "Darwin" in psys:
                osTarget = 'macos-10.12'
            elif "Linux" in psys:
                osTarget = 'ubuntu-14.04'
            else:
                raise NotImplementedException()
        elif osName == 'nt' and "Windows" in psys:
            osTarget = 'windows'
        else:
            raise NotImplementedException()
        return getSupportFile(os.path.join('support', 'unison', osTarget))



import subprocess
import os
import random
import string
import urllib.request, urllib.parse, urllib.error
try:
    import readline  # Make raw_input nicer to use, but don't make it required
    readline
except ImportError:
    pass  # Optional dependency
from substance import (Command)
from substance.logs import (logging)
from substance.config import (Config)

logger = logging.getLogger(__name__)


class Hatch(Command):

    def getUsage(self):
        return "substance hatch [options] TEMPLATE [REVISION]"

    def getHelpTitle(self):
        return "Create a new project based on a git or tarball template."

    def getHelpDetails(self):
        return "\n".join([
            "Arguments:",
            "  TEMPLATE     Git repository, or short name. Examples of valid TEMPLATEs:",
            "               php-project        (resolves to https://github.com/turbulent/template-php-project)",
            "               bob/my-template    (resolves to https://github.com/bob/my-template)",
            "               ssh://git@gitlab.turbulent.ca:6666/templates/heap-project.git",
            "               https://www.example.com/path/to/tarball.tar.gz",
            "               file:///home/bob/my-git-template-directory",
            "",
            "  REVISION     Refers to the git ref of the template repository. Defaults to 'master'."
        ])

    def getShellOptions(self, optparser):
        optparser.add_option("-s", "--strip", dest="strip", default="1",
                             help="Strip X number of leading components from tarball paths", action="store_true")
        return optparser

    def main(self):
        tpl = self.getArg(0)
        if not tpl:
            return self.exitError("You MUST provide a template name or git repository URL!")
        ref = self.getArg(1) or "master"

        target_dir = "".join(tpl.split('/')[-1:]).split('.')[0]

        if not tpl.startswith('ssh://') and not tpl.startswith('https://') and not tpl.startswith('file://'):
            splt = tpl.split("/", 2)
            if len(splt) == 2:
                tpl = 'https://github.com/%s/%s/archive/%s.tar.gz' % (
                    splt[0], splt[1], ref)
                target_dir = splt[1]
            else:
                target_dir = tpl
                tpl = 'https://github.com/turbulent/template-%s/archive/%s.tar.gz' % (
                    tpl, ref)
        install_dir = os.getcwd() + ("/%s" % target_dir)
        if os.path.exists(install_dir):
            print("\n!!! Install directory exists! Make sure it's empty or Some files may be overwritten !!!\n")

        print("You are about to hatch a new project.")
        print("  Template used: %s" % tpl)
        print("  Ref (version): %s" % ref)
        if not tpl.endswith('.tar.gz'):
            print("  Ref (version): %s" % ref)
        print("  Path         : %s" % install_dir)
        print("")

        if not self.confirm("Are you SURE you wish to proceed?"):
            return self.exitOK("Come back when you've made up your mind!")

        if not os.path.exists(install_dir):
            os.mkdir(target_dir, 0o755)
        os.chdir(target_dir)
        print("Switching to directory: %s" % install_dir)
        print("Downloading template archive...")
        if tpl.endswith('.tar.gz'):
            # With tar archives, everything is usually packaged in a single directory at root of archive
            strip = int(self.getOption('strip'))
            urllib.request.urlretrieve(tpl, 'tpl.tar.gz')
        else:
            strip = 0  # git archive never packages in a single root directory
            if self.proc(['git', 'archive', '--verbose', '-o', 'tpl.tar.gz', '--remote=' + tpl, ref]):
                return self.exitError('Could not download template %s@%s!' % (tpl, ref))

        print("Extracting template archive...")
        if self.proc(['tar', '--strip', str(strip), '-xf', 'tpl.tar.gz']):
            return self.exitError('Could not extract template archive!')

        # Acquire list of files
        print("Getting list of files in template...")
        out = subprocess.check_output(['tar', '-tf', 'tpl.tar.gz'], universal_newlines=True)
        tplFiles = ['/'.join(l.split('/')[strip:]) for l in out.split('\n') if l]
        tplFiles = [l for l in tplFiles if os.path.isfile(l)]

        print("Cleaning up template archive...")
        if self.proc(['rm', 'tpl.tar.gz']):
            return self.exitError('Could not unlink temporary template archive!')

        # At this point, we have all the files we need
        hatchfile = os.path.join('.substance', 'hatch.yml')
        if os.path.isfile(hatchfile):
            config = Config(hatchfile)
            res = config.loadConfigFile()
            if res.isFail():
                return self.exitError("Could not open %s for reading: %s" % (hatchfile, res.getError()))

            # Execute pre-script if any
            for cmd in config.get('pre-script', []):
                print(cmd)
                subprocess.call(cmd, shell=True)

            vardefs = config.get('vars', {})
            # Autogenerate a secret
            chars = string.ascii_lowercase + string.ascii_uppercase + string.digits
            variables = {
                '%hatch_secret%': ''.join(random.SystemRandom().choice(chars) for _ in range(32))
            }
            if vardefs:
                print("This project has variables. You will now be prompted to enter values for each variable.")
                for varname in vardefs:
                    val = ''
                    required = vardefs[varname].get('required', False)
                    default = vardefs[varname].get('default', '')
                    description = vardefs[varname].get('description', '')
                    while not val:
                        val = input("%s (%s) [%s]: " % (
                            varname, description, default))
                        if default and not val:
                            val = default
                        if not required:
                            break
                    variables[varname] = val

            summary = "\n".join(["  %s: %s" % (k, variables[k])
                                 for k in list(variables.keys())])
            print("Summary: ")
            print(summary)
            if not self.confirm("OK to replace tokens?"):
                return self.exitOK("Operation aborted.")

            print("Replacing tokens in files...")
            sed = "; ".join(["s/%s/%s/g" % (k, variables[k].replace('/', '\\/'))
                             for k in list(variables.keys())])
            for tplFile in tplFiles:
                if self.proc(['sed', '-i.orig', sed, tplFile]):
                    return self.exitError("Could not replace variables in files!")
                bakFile = tplFile + ".orig"
                if os.path.isfile(bakFile):
                    if self.proc(['rm', bakFile]):
                        logger.warn(
                            "Could not unlink backup file %s; you may have to remove it manually.", bakFile)

            # Execute post-script if any
            for cmd in config.get('post-script', []):
                print(cmd)
                subprocess.call(cmd, shell=True)

            # Remove hatchfile
            if self.proc(['rm', hatchfile]):
                return self.exitError('Could not unlink %s!' % hatchfile)

        print("Project hatched!")
        return self.exitOK()

    def proc(self, cmd, variables=None):
        logger.debug(subprocess.list2cmdline(cmd))
        return subprocess.call(cmd)

    def confirm(self, prompt):
        confirm = ''
        while confirm not in ('Y', 'y'):
            confirm = input("%s (Y/n) " % prompt) or 'Y'
            if confirm in ('N', 'n'):
                return False
        return True



from substance.monads import *
from substance.logs import *
from substance import (Command)
from tabulate import tabulate

logger = logging.getLogger(__name__)


class Ls(Command):
    def getUsage(self):
        return "substance box ls [options]"

    def getHelpTitle(self):
        return "List substance box available locally"

    def main(self):
        self.core.getBoxes() \
            .mapM(self.tabulateBox) \
            .bind(self.tabulateBoxes) \
            .catch(self.exitError) \
            .bind(logger.info)

    def tabulateBox(self, box):
        return OK([box.namespace, box.name, box.version, box.registry, box.archiveSHA1])

    def tabulateBoxes(self, boxes):
        headers = ["NAMESPACE", "NAME", "VERSION", "REGISTRY", "HASH"]
        table = tabulate(boxes, headers=headers, tablefmt="plain")
        return OK(table)



from substance import (Engine, Command)
from substance.exceptions import (LinkCommandError)


class Fallback(Command):

    def getUsage(self):
        return ""

    def getHelpTitle(self):
        return ""

    def getShellOptions(self, optparser):
        return optparser

    # Disable parsing for this command
    def execute(self, args=None):
        self.input = args
        self.args = args

        self.main()

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.envLoadCurrent) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.loadSubenvConfigFile) \
            .bind(Engine.envExecAlias, alias=self.name, args=self.args) \
            .catch(self.onEngineException)

    def onEngineException(self, err):
        if isinstance(err, LinkCommandError):
            # Hide message, forward status code instead
            return self.exitError(None, code=err.code)
        return self.exitError(err)



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from substance.exceptions import (SubstanceError)


class Suspend(Command):
    def getUsage(self):
        return "substance suspend [ENGINE NAME]"

    def getHelpTitle(self):
        return "Suspend an engine gracefully."

    def main(self):
        name = self.getInputName()

        self.core.loadEngine(name) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.suspend) \
            .catch(self.exitError)



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate
from substance.constants import (EngineStates)
from substance.exceptions import (EngineNotRunning)

logger = logging.getLogger(__name__)


class Status(Command):
    def getShellOptions(self, optparser):
        return optparser

    def getUsage(self):
        return "substance status [options]"

    def getHelpTitle(self):
        return "Show substance engine and env status"

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.envLoadCurrent) \
            .bind(Engine.envStatus, full=True) \
            .bind(self.printStatus) \
            .catch(self.exitError)

    def printStatus(self, envStatus):
        engine = envStatus.get('engine')
        containers = envStatus.get('containers')

        headers = ["ENGINE", "SUBENV"]
        cols = [[engine.name, engine.currentEnv]]
        table = tabulate(cols, headers=headers, tablefmt="plain")

        logger.info(table)
        if containers:
            logger.info("")
            logger.info(containers)
        return OK(None)



import re
import os
from collections import OrderedDict
from substance.monads import *
from substance.logs import *
from substance.constants import Constants
from substance.exceptions import (MachineDoesNotExist, SubstanceDriverError)
from .vbox import (vboxManager, _vboxLineEnding)
from .exceptions import *
from substance.shell import Shell
from substance.path import (inner, outer)
from functools import reduce

logger = logging.getLogger(__name__)

MachineStates = Constants(
    POWEROFF="poweroff",
    SAVED="saved",
    ABORTED="aborted",
    PAUSED="paused",
    STUCK="stuck",
    RESTORING="restoring",
    SNAPSHOTTING="snapshotting",
    SETTING_UP="setting up",
    ONLINE_SNAPSHOTTING="online snapshotting",
    RESTORING_SNAPSHOT="restoring snapshot",
    DELETING_SNAPSHOT="deleting snapshot",
    LIVE_SNAPSHOTTING="live snapshotting",
    RUNNING="running",
    STARTING="starting",
    STOPPING="stopping",
    SAVING="saving",
    UNKNOWN="unknown",
    INACCESSIBLE="inaccessible",
    INEXISTENT="inexistent"
)

AdapterTypes = Constants(
    NONE="none",
    NULL="null",
    NAT="nat",
    HOSTONLY="hostonly",
    NATNET="natnet",
    INTNET="intnet",
    BRIDGED="bridged"
)


class AdapterSettings(object):
    def __init__(self, type, attachedTo, mac=None, promiscuous=False):
        self.type = type
        self.attachedTo = attachedTo
        self.mac = mac
        self.promiscuous = promiscuous

    def getTypeArg(self):
        return {
            AdapterTypes.NAT: 'natnet',
            AdapterTypes.HOSTONLY: 'hostonlyadapter',
            AdapterTypes.NATNET: 'nat-network',
            AdapterTypes.BRIDGED: 'bridgeadapter',
            AdapterTypes.INTNET: 'intnet'
        }.get(self.type, 'natnet')

    def getAsArg(self, nicId=0):
        vals = self.__dict__.copy()
        vals['nicId'] = nicId
        vals['typeArg'] = self.getTypeArg()
        return "--nic%(nicId)d %(type)s --%(typeArg)s%(nicId)d \"%(attachedTo)s\"" % vals

    def __repr__(self):
        return "Adapter(nic=%(nic)s,nictype=%(nictype)s,mac=%(mac)s,promiscuous=%(promiscuous)s" % self.__dict__


class SharedFolder(object):
    def __init__(self, name, hostPath, vboxName=None):
        self.name = name
        self.hostPath = hostPath
        self.vboxName = vboxName

    def __repr__(self):
        return "SharedFolder(%(name)s -> %(hostPath)s, %(vboxName)s)" % self.__dict__

# -- Import API


def inspectOVF(ovfFile):
    '''
    Inspect an OVF file to extract it's examined output
    '''
    return vboxManager("import", '-n "%s"' % outer(ovfFile))


def makeImportParams(inspection, name, engineProfile=None):
    '''
    Make the import parameter commands based on name, engine profile and inspection
    '''
    suggestedName = re.search(r'Suggested VM name "(.+?)"', inspection)
    if not suggestedName:
        return Fail(SubstanceDriverError("Invalid OVF: File contains no VM Name"))
    else:
        suggestedName = suggestedName.group(1)

    suggestedPath = re.search(r'Suggested VM base folder "(.+?)"', inspection)
    if not suggestedPath:
        return Fail(SubstanceDriverError("Invalid OVF: File contains no VM Suggested Path"))
    else:
        suggestedPath = suggestedPath.group(1)

    logger.debug("OVF Suggested VM Name \"%s\"", suggestedName)
    logger.debug("OVF Suggested VM base folder \"%s\"", suggestedPath)

    importParams = ["--vsys 0 --vmname \"%s\"" % name]
    if engineProfile:
        importParams.append("--cpus %d" % engineProfile.cpus)
        importParams.append("--memory %d" % engineProfile.memory)

    diskScan = re.compile(
        r'(\d+): Hard disk image: source image=.+, target path=(.+),')
    for disk in diskScan.findall(inspection):
        diskType = os.path.splitext(disk[1])[1]
        diskPath = outer(os.path.join(suggestedPath, '%s%s' % (name, diskType)))
        importParams.append("--unit %s --disk \"%s\"" % (disk[0], diskPath))

    return OK(importParams)


def importOVF(importParams, name, ovfFile):
    '''
    Import the OVF file as a virtual box vm.
    '''
    ovfFile = outer(ovfFile)
    importParams.insert(0, '"' + ovfFile + '"')
    return vboxManager("import", " ".join(importParams)) \
        .then(defer(readMachineID, name))

# -- Read functions


def readMachines():
    return vboxManager("list", "vms").bind(parseMachinesList)


def readMachineID(name):
    '''
    Retrieve the driver specific machine ID for a machine name.
    '''
    def findMachineID(machs):
        revmach = dict(list(zip(list(machs.values()), list(machs.keys()))))
        return OK(revmach[name]) if name in revmach else Fail(MachineDoesNotExist("No Machine ID found for \"%s\"" % name))

    return vboxManager("list", "vms") \
        .bind(parseMachinesList) \
        >> findMachineID


def readMachineInfo(uuid):
    return vboxManager("showvminfo", "--machinereadable \"%s\"" % uuid) \
        .bind(parseMachineInfo)


def readMachineExists(uuid):
    return vboxManager("list", "vms").bind(defer(parseMachinesForID, uuid=uuid))


def readMachineState(uuid):
    return vboxManager("showvminfo", "--machinereadable \"%s\"" % uuid) \
        .catch(lambda err: OK("inexistent") if err.code == "VBOX_E_OBJECT_NOT_FOUND" else Fail(err)) \
        .bind(parseMachineState)


def readWaitGuestProperty(uuid, pattern, timeout=1000):
    logger.debug("Read wait(%ss) on pattern %s on %s" %
                 (timeout, pattern, uuid))
    return vboxManager("guestproperty", "wait %s %s --timeout %s" % (uuid, pattern, timeout))


def readGuestProperty(uuid, prop):
    logger.debug("Read guest property %s on %s" % (prop, uuid))
    return vboxManager("guestproperty", "get %s %s" % (uuid, prop)) \
        .bind(parseGuestProperty)


def readGuestAddVersion(uuid):
    return readGuestProperty(uuid, "/VirtualBox/GuestAdd/Version") \
        .bind(parseGuestAddVersion)

# -- Parsing functions


def parseMachinesList(vms):
    '''
    Parse the output of "list vms" and return a dict of machine name to machine id.
    '''
    matcher = re.compile(r'"([^"]+)" {([^}]+)}')
    machines = {}
    for line in vms.split(_vboxLineEnding()):
        parts = matcher.match(line)
        if parts:
            machines[parts.group(2)] = parts.group(1)
    return OK(machines)


def parseMachinesForID(vms, uuid):
    '''
    Parse the output of list vms to find if uuid exists.
    '''
    if re.search(r'"([^"]+)" {' + re.escape(uuid) + '}', vms, re.M):
        return OK(True)
    else:
        return OK(False)


def parseMachineInfo(machInfo):
    '''
    Parse Virtual Box machine info into a dict.
    '''
    lines = machInfo.split(_vboxLineEnding())
    machDict = OrderedDict()
    for line in lines:
        try:
            var, value = line.split("=", 1)
            value = value.strip('"')
            machDict[var] = value
        except ValueError:
            pass
    return OK(machDict)


def parseMachineState(vminfo):
    '''
    Parse the output of showvminfo to extract the VM state
    '''
    if re.search(r'^name="<inaccessible>"\r?\n', vminfo, re.MULTILINE):
        return OK('inaccessible')

    stateMatch = re.search(r'^VMState="(.+?)"\r?\n', vminfo, re.MULTILINE)
    if stateMatch:
        return OK(stateMatch.group(1))
    else:
        return OK('inexistent')


def parseGuestProperty(prop):
    prop = prop.strip()

    if prop == 'No value set!' or not prop:
        return OK(None)

    match = re.match(r'^Value: (.+?)$', prop)
    logger.debug("Guest Property: %s" % prop)
    if match:
        return OK(match.group(1))
    else:
        return Fail(VirtualBoxError(None, "Failed to to parse guest property output : %s" % prop))


def parseGuestAddVersion(guestAdd):
    if guestAdd is None:
        return Fail(VirtualBoxMissingAdditions("VirtualBox guest additions are not installed."))

    parts = guestAdd.split("_", 1)
    if len(parts) >= 1:
        return OK(parts[0])
    else:
        return Fail(VirtualBoxMissingAdditions("VirtualBox guest additions are not installed."))


def parseSharedFolders(machInfo):
    def extractFolders(acc, k):
        match = re.match(r'^SharedFolderNameMachineMapping(\d+)', k)
        if match:
            idx = match.group(1)
            acc.append(SharedFolder(
                name=machInfo[k], hostPath=machInfo["SharedFolderPathMachineMapping%s" % idx], vboxName=k))
        return acc

    return OK(reduce(extractFolders, list(machInfo.keys()), []))

# -- Modify


def configureAdapter(uuid, adapterId, adapterSettings):
    return vboxManager("modifyvm", "\"%s\" %s" % (uuid, adapterSettings.getAsArg(adapterId)))


def configureProfile(uuid, cpus, memory):
    return vboxManager("modifyvm", "\"%s\" --cpus %s --memory %s" % (uuid, cpus, memory))

# -- Control


def start(uuid):
    return vboxManager("startvm", "--type headless \"%s\"" % (uuid)) \
        .bind(lambda x: OK(uuid))


def halt(uuid):
    return vboxManager("controlvm", "\"%s\" poweroff soft" % uuid) \
        .bind(lambda x: OK(uuid))


def suspend(uuid):
    return vboxManager("controlvm", "\"%s\" savestate" % uuid) \
        .bind(lambda x: OK(uuid))


def pause(uuid):
    return vboxManager("controlvm", "\"%s\" pause" % uuid) \
        .bind(lambda x: OK(uuid))


def resume(uuid):
    return vboxManager("controlvm", "\"%s\" resume" % uuid) \
        .bind(lambda x: OK(uuid))


def terminate(uuid):
    return vboxManager("controlvm", "\"%s\" poweroff" % uuid) \
        .bind(lambda x: OK(uuid))


def delete(uuid):
    return vboxManager("unregistervm", "--delete \"%s\"" % (uuid)) \
        .bind(lambda x: OK(uuid))

# -- Shared Folders


def addSharedFolder(folder, uuid):
    return vboxManager("sharedfolder", "add \"%s\" --name \"%s\" --hostpath \"%s\"" % (uuid, folder.name, folder.hostPath)) \
        .then(defer(enableSharedFolderSymlinks, folder=folder, uuid=uuid))


def enableSharedFolderSymlinks(folder, uuid):
    symlinksKey = "VBoxInternal2/SharedFoldersEnableSymlinksCreate/%s" % folder.name
    return vboxManager("setextradata", "\"%s\" \"%s\" \"%s\"" % (uuid, symlinksKey, 1))


def removeSharedFolder(folder, uuid):
    return vboxManager("sharedfolder", "remove \"%s\" --name \"%s\"" % (uuid, folder.name))


def addSharedFolders(folders, uuid):
    return OK(list(map(defer(addSharedFolder, uuid=uuid), folders)))


def removeSharedFolders(folders, uuid):
    return OK(list(map(defer(removeSharedFolder, uuid=uuid), folders)))


def clearSharedFolders(uuid):
    return readSharedFolders(uuid) \
        .bind(removeSharedFolders, uuid=uuid)


def readSharedFolders(uuid):
    return readMachineInfo(uuid) \
        .bind(parseSharedFolders)




import sys
import logging
from collections import OrderedDict
from substance import (SubProgram, Core)


class Engine(SubProgram):

    def __init__(self):
        super(Engine, self).__init__()

    def setupCommands(self):
        self.addCommand('ls', 'substance.command.engine.ls')
        self.addCommand('create', 'substance.command.engine.create')
        self.addCommand('delete', 'substance.command.engine.delete')
        self.addCommand('launch', 'substance.command.engine.launch')
        self.addCommand('halt', 'substance.command.engine.halt')
        self.addCommand('suspend', 'substance.command.engine.suspend')
        self.addCommand('deprovision', 'substance.command.engine.deprovision')
        self.addCommand('edit', 'substance.command.engine.edit')
        self.addCommand('ssh', 'substance.command.engine.ssh')
        self.addCommand('env', 'substance.command.engine.env')
        self.addCommand('sshinfo', 'substance.command.engine.sshinfo')
        return self

    def getShellOptions(self, optparser):
        return optparser

    def getUsage(self):
        return "substance engine [options] COMMAND [command-options]"

    def getHelpTitle(self):
        return "Substance engine management"

    def initCommand(self, command):
        command.core = self.core
        return command



import os
import logging
from substance.logs import *
from substance.monads import *
from substance import (Engine, Command, Shell)
from substance.exceptions import (EngineNotFoundError)


class Edit(Command):

    def getUsage(self):
        return "substance engine edit [ENGINE NAME]"

    def getHelpTitle(self):
        return "Edit an engine configuration"

    def main(self):

        name = self.getInputName()

        self.core.loadEngine(name=name) \
            .bind(self.editEngineConfig) \
            .bind(Engine.loadConfigFile) \
            .then(dinfo("Engine \"%s\" configuration was updated.", name)) \
            .catch(self.exitError)

    def editEngineConfig(self, engine):
        cmd = os.environ.get('EDITOR', 'vi') + ' ' + \
            engine.config.getConfigFile()
        return Shell.call(cmd).map(lambda x: engine)



import os
import logging
from substance.monads import *
from substance.shell import Shell
from substance.utils import (readYAML, writeYAML, mergeDict)
from substance.exceptions import (
    FileSystemError, FileDoesNotExist, ConfigValidationError)

from collections import namedtuple
from functools import reduce

logger = logging.getLogger(__name__)


class Config(object):

    def __init__(self, configFile=None):
        self.loaded = False
        self.config = {}
        self.configFile = configFile

    def getConfigFile(self):
        return self.configFile

    def set(self, key, value):
        self.config[key] = value
        return value

    def get(self, key, default=None):
        return self.config.get(key, default)

    def getBlockKey(self, block, key, default=None):
        return self.config.get(block, {}).get(key, default)

    def getConfig(self):
        return self.config

    def setConfig(self, config):
        mergeDict(self.config, config)
        return OK(self)

    def saveConfig(self, config=None):
        config = self.config if not config else config
        return Try.attempt(writeYAML, self.configFile, config)

    def readConfigFile(self):
        return Try.attempt(readYAML, self.configFile)

    def loadConfigFile(self):
        if not os.path.exists(self.configFile):
            return Fail(FileDoesNotExist("File does not exist: %s" % self.configFile))
        logger.debug("Loading config file: %s", self.configFile)
        return self.readConfigFile() >> self.setConfig

    def validateFieldsPresent(self, block, fields):
        def reducer(acc, x): return Fail(ConfigValidationError(
            "Property does not exist in configuration: %s" % x)) if x not in block else OK(None)
        return reduce(reducer, fields, OK(None))



import os
import socket
import logging
from contextlib import closing
from collections import (OrderedDict, namedtuple)

import substance.core
from substance.monads import *
from substance.config import (Config)
from substance.logs import *
from substance.shell import Shell
from substance.link import Link
from substance.box import Box
from substance.utils import mergeDict, mergeDictOverwrite, parseDotEnv, CommandList, Memoized
from substance.path import (getHomeDirectory, inner)
from substance.hosts import SubHosts
from substance.driver.virtualbox import VirtualBoxDriver
from substance.constants import (EngineStates, DefaultEngineBox, Orchestrators)
from substance.platform import (isWithinWindowsSubsystem)
from substance.orchestrator import (Dockwrkr, Compose)

from substance.exceptions import (
    FileSystemError,
    ConfigValidationError,
    EngineAlreadyRunning,
    EngineNotRunning,
    EngineExistsError,
    EngineNotProvisioned,
    EngineProvisioned,
    EnvNotDefinedError,
    InvalidCommandError,
    InvalidOrchestrator
)
from substance.syncher.unison import UnisonSyncher

logger = logging.getLogger(__name__)


class EngineProfile(object):
    def __init__(self, cpus=2, memory=1024):
        self.cpus = cpus
        self.memory = memory

    def __repr__(self):
        return "%s" % {'cpus': self.cpus, 'memory': self.memory}


class EngineFolder(object):
    def __init__(self, name, mode, hostPath, guestPath, uid=None, gid=None, umask=None, excludes=[], syncArgs=[]):
        self.name = name
        self.mode = mode
        self.hostPath = hostPath
        self.guestPath = guestPath
        self.uid = uid if uid else 1000
        self.gid = gid if gid else 1000
        self.umask = umask if umask else "0022"
        self.excludes = excludes
        self.syncArgs = syncArgs

    def setExcludes(self, exs=[]):
        self.excludes = exs

    def getExcludes(self):
        return self.excludes

    def __eq__(self, other):
        return True if self.__hash__() == other.__hash__() else False

    def __hash__(self):
        return hash((self.name, self.hostPath, self.guestPath))

    def __repr__(self):
        return "EngineFolder(name=%(name)s, %(mode)s host:%(hostPath)s -> guest:%(guestPath)s)" % self.__dict__


class Engine(object):

    def __init__(self, name, enginePath, core):
        self.name = name
        self.enginePath = enginePath
        self.core = core
        self.link = None
        configFile = os.path.join(self.enginePath, "engine.yml")
        self.config = Config(configFile)
        self.logAdapter = EngineLogAdapter(logger, self.__dict__)
        self.currentEnv = None
        self.subenvConfig = None

    def getDefaults(self):
        defaults = OrderedDict()
        defaults['name'] = 'default'
        defaults['driver'] = 'virtualbox'
        defaults['id'] = None
        defaults['box'] = DefaultEngineBox
        defaults['profile'] = EngineProfile().__dict__
        defaults['docker'] = OrderedDict()
        defaults['docker']['port'] = 2375
        defaults['docker']['tls'] = False
        defaults['docker']['certificateFile'] = None
        defaults['network'] = OrderedDict()
        defaults['network']['privateIP'] = None
        defaults['network']['publicIP'] = None
        defaults['network']['sshIP'] = None
        defaults['network']['sshPort'] = 4500
        defaults['devroot'] = OrderedDict()
        defaults['devroot']['path'] = getHomeDirectory('devroot')
        defaults['devroot']['mode'] = 'unison'
        defaults['devroot']['syncArgs'] = [
            '-ignore', 'Path */var',
            '-ignore', 'Path */data'
        ]
        defaults['devroot']['excludes'] = [
            '{.*,*}.sw[pon]',
            '.bash*',
            '.composer',
            '.git',
            '.idea',
            '.npm',
            '.ssh',
            '.viminfo'
        ]
        aliases = ['make', 'composer', 'npm', 'heap', 'watch']
        defaults['aliases'] = {alias: {'container': 'web', 'user': 'heap',
                                       'cwd': '/vol/website', 'args': [alias]} for alias in aliases}
        defaults['aliases']['watch']['container'] = 'devs'
        defaults['aliases']['heap']['args'][0] = 'vendor/bin/heap'

        return defaults

    def validateConfig(self, config):
        fields = ['name', 'driver', 'profile',
                  'docker', 'network', 'devroot', 'box']

        ops = []
        ops.append(self.config.validateFieldsPresent(
            config.getConfig(), fields))
        ops.append(self.confValidateName(config.get('name', None)))
        ops.append(self.core.validateDriver(config.get('driver', None)))
        ops.append(self.confValidateDevroot(config.get('devroot', {})))

        def dd(err):
            print(("%s" % err))
            return ConfigValidationError(err.message)

        return Try.sequence(ops) \
            .catch(dd) \
            .then(lambda: OK(config.getConfig()))

    def confValidateName(self, name):
        if name != self.name:
            return Fail(ConfigValidationError("Invalid name property in configuration (got %s expected %s)" % (config.get('name'), self.name)))
        return OK(name)

    def confValidateDevroot(self, devroot):
        path = devroot.get('path', '')
        path = os.path.expanduser(path)
        if not path:
            return Fail(ConfigValidationError("devroot path configuration is missing."))
        elif not os.path.isdir(path):
            self.logAdapter.info(
                "WARNING: devroot '%s' does not exist locally." % path)

        mode = devroot.get('mode', None)
        if not mode:
            return Fail(ConfigValidationError("devroot mode is not set."))
        elif mode not in ['sharedfolder', 'rsync', 'unison']:
            # XXX Fix hardcoded values.
            return Fail(ConfigValidationError("devroot mode '%s' is not valid." % mode))

        return OK(devroot)

    def getName(self):
        return self.name

    def getID(self):
        return self.config.id

    def getEnginePath(self):
        return self.enginePath

    def getEngineFilePath(self, fileName):
        return os.path.join(self.enginePath, fileName)

    def getDockerURL(self):
        if self.getDriverID():
            return "tcp://%s:%s" % (self.getPublicIP(), self.getDockerPort())

    def getDockerPort(self):
        return self.config.get('docker', {}).get('port', 2375)

    def getDockerEnv(self):
        env = OrderedDict()
        # env['DOCKER_API_VERSION'] = '1.19'
        env['DOCKER_HOST'] = self.getDockerURL()
        env['DOCKER_TLS_VERIFY'] = ''
        return env

    def getPublicIP(self):
        return self.config.get('network').get('publicIP', None)

    def getPrivateIP(self):
        return self.config.get('network').get('privateIP', None)

    def getDriver(self):
        return self.core.getDriver(self.config.get('driver'))

    def getEngineProfile(self):
        profile = self.config.get('profile', {})
        return EngineProfile(**profile)

    def getDriverID(self):
        return self.config.get('id', None)

    def setDriverID(self, driverID):
        self.config.set('id', driverID)
        return OK(self)

    def getSSHPort(self):
        return self.config.getBlockKey('network', 'sshPort', 22)

    def getSSHIP(self):
        return self.config.getBlockKey('network', 'sshIP', 'localhost')

    def getConnectInfo(self):
        info = {
            'hostname': self.getSSHIP(),
            'port': self.getSSHPort()
        }
        return info

    def getDNSName(self):
        tld = self.core.config.get('tld')
        return self.name + tld

    def getSyncher(self):
        syncMode = self.config.get('devroot').get('mode')
        keyfile = self.core.getInsecureKeyFile()
        return UnisonSyncher(engine=self, keyfile=keyfile)

    def clearDriverID(self):
        self.config.set('id', None)
        return OK(self)

    def loadConfigFile(self):
        return self.config.loadConfigFile().bind(self.validateConfig).map(self.chainSelf)

    def loadSubenvConfigFile(self):
        basePath = os.path.expanduser(self.config.get('devroot').get('path'))
        self.subenvConfig = Config(os.path.join(basePath,
                                   '%s/.substance/subenv.yml' % self.currentEnv))
        return self.subenvConfig.loadConfigFile().map(self.chainSelf)

    def loadState(self):
        ddebug("Engine load state %s" % self.name)
        return self.fetchState().bind(self.setState).map(self.chainSelf)

    def setState(self, state):
        self.state = state
        return OK(None)

    def create(self, config=None, profile=None):
        if os.path.isdir(self.enginePath):
            return Fail(EngineExistsError("Engine \"%s\" already exists." % self.name))

        return Shell.makeDirectory(self.enginePath) \
            .then(defer(self.makeDefaultConfig, config=config, profile=profile)) \
            .bind(self.__ensureDevroot) \
            .bind(self.validateConfig) \
            .then(self.config.saveConfig) \
            .bind(dinfo("Generated default engine configuration in %s", self.config.configFile)) \
            .map(self.chainSelf)

    def __ensureDevroot(self, config):
        devroot = os.path.expanduser(config.get('devroot', {}).get('path'))
        if not os.path.isdir(devroot):
            self.logAdapter.info("Creating devroot at %s" % devroot)
            return Shell.makeDirectory(devroot).then(lambda: OK(config))
        return OK(config)

    def remove(self):
        if not os.path.isdir(self.enginePath):
            return Fail(FileSystemError("Engine \"%s\" does not exist." % self.name))
        if self.isProvisioned():
            return Fail(EngineProvisioned("Engine \"%s\" is provisioned." % self.name))
        return Shell.nukeDirectory(self.enginePath)

    def makeDefaultConfig(self, config=None, profile=None):
        default = self.getDefaults()
        default["name"] = self.name
        default["devroot"]["path"] = os.path.join(
            self.core.config.get('devroot'), self.name)

        cf = {}
        mergeDict(cf, default)
        mergeDict(cf, config)

        if profile:
            cf['profile']['cpus'] = profile.cpus
            cf['profile']['memory'] = profile.memory

        for ck, cv in cf.items():
            self.config.set(ck, cv)

        return OK(self.config)

    def getProfile(self):
        prof = self.config.get('profile', {})
        return EngineProfile(prof.get('cpus'), prof.get('memory'))

    def isProvisioned(self):
        '''
        Check that this engine has an attached Virtual Machine.
        '''
        return True if self.getDriverID() else False

    def validateProvision(self):
        '''
        Check that the provision attached to this engine is valid in the driver.
        '''
        if not self.isProvisioned():
            return OK(False)

        machID = self.getDriverID()
        return self.getDriver().exists(machID)

    def isRunning(self):
        return self.fetchState().map(lambda state: (state is EngineStates.RUNNING))

    def isSuspended(self):
        return self.fetchState().map(lambda state: (state is EngineStates.SUSPENDED))

    def fetchState(self):
        if not self.isProvisioned():
            return OK(EngineStates.INEXISTENT)
        return self.getDriver().getMachineState(self.getDriverID())

    def launch(self):
        '''
        # 1. Check that we know about a provisioned machined for this engine. Provision if not.
        # 2. Validate that the machine we know about is still provisioned. Provision if not.
        # 3. Check the machine state, boot accordingly
        # 4. Setup guest networking
        # 4. Fetch the guest IP from the machine and store it in the machine state.
        '''
        return self.provision() \
            .then(self.start) \
            .then(self.updateNetworkInfo) \
            .then(self.addToHostsFile) \
            .then(self.envLoadCurrent) \
            .catchError(EnvNotDefinedError, lambda e: OK(None)) \
            .then(dinfo("Engine \"%s\" has been launched.", self.name)) \
            .then(self.envStart if self.currentEnv else lambda: OK(None))

    def addToHostsFile(self):
        self.logAdapter.info(
            "Registering engine as '%s' in local hosts file" % self.getDNSName())
        return SubHosts.register(self.getDNSName(), self.getPublicIP())

    def removeFromHostsFile(self):
        self.logAdapter.info("Removing engine from local hosts file")
        return SubHosts.unregister(self.getDNSName())

    def updateNetworkInfo(self):
        self.logAdapter.info("Updating network info from driver")
        return self.__readDriverNetworkInfo().bind(self.saveDriverNetworkInfo)

    def postLaunch(self):
        return self.setHostname() \
            .then(self.mountFolders)

    def saveDriverNetworkInfo(self, info):
        self.logAdapter.debug("Network information for machine: %s" % info)
        net = self.config.get('network', OrderedDict())
        net.update(info)
        self.config.set('network', net)
        return self.config.saveConfig()

    def start(self):

        state = self.fetchState()
        if state.isFail():
            return state

        if state.getOK() is EngineStates.RUNNING:
            return EngineAlreadyRunning("Engine \"%s\" is already running" % self.name)

        if state.getOK() is EngineStates.SUSPENDED:
            return self.__start().then(self.__waitForReady)
        else:
            return self.__configure() \
                .then(self.updateNetworkInfo) \
                .then(self.__start) \
                .then(self.__waitForReady) \
                .then(self.__waitForNetwork) \
                .then(self.postLaunch)

    def readBox(self):
        return self.core.readBox(self.config.get('box'))

    def provision(self):

        logger.info("Provisioning engine \"%s\" with driver \"%s\"",
                    self.name, self.config.get('driver'))

        prov = self.validateProvision()
        if prov.isFail():
            return prov
        elif prov.getOK():
            return OK(self)

        box = self.readBox().bind(Box.fetch)
        if box.isFail():
            return box
        box = box.getOK()

        return self.getDriver().importMachine(self.name, box.getOVFFile(), self.getEngineProfile()) \
            .bind(self.setDriverID) \
            .then(self.config.saveConfig) \
            .map(self.chainSelf)
        # .then(self.__configure) \
        # .then(self.updateNetworkInfo) \

    def deprovision(self):

        if not self.isProvisioned():
            return Fail(EngineNotProvisioned("Engine \"%s\" is not provisioned." % self.name, self))

        self.__cacheLink(None)

        return self.validateProvision() \
            .thenIfTrue(self.isRunning) \
            .thenIfTrue(self.__terminate) \
            .then(self.__delete) \
            .then(self.clearDriverID)  \
            .then(self.config.saveConfig)  \
            .then(self.removeFromHostsFile) \
            .then(dinfo("Engine \"%s\" has been deprovisioned.", self.name)) \
            .map(self.chainSelf)

    def suspend(self):
        if not self.isProvisioned():
            return Fail(EngineNotProvisioned("Engine \"%s\" is not provisioned." % self.name))
        self.__cacheLink(None)

        return self.validateProvision() \
            .thenIfTrue(self.isRunning) \
            .thenIfFalse(failWith(EngineNotRunning("Engine \"%s\" is not running." % self.name))) \
            .then(self.__suspend) \
            .then(dinfo("Engine \"%s\" has been suspended.", self.name)) \
            .map(chainSelf)

        # XXX Insert wait for suspension

    def stop(self, force=False):

        self.__cacheLink(None)

        operation = self.validateProvision() \
            .thenIfTrue(self.isRunning) \
            .thenIfFalse(failWith(EngineNotRunning("Engine \"%s\" is not running." % self.name)))

        if(force):
            operation >> self.__terminate >> dinfo(
                "Engine \"%s\" has been terminated.", self.name)
        else:
            operation >> self.__halt >> dinfo(
                "Engine \"%s\" has been halted.", self.name)

        # XXX insert wait for stopped state
        return operation.map(self.chainSelf)

    def readLink(self):
        if self.link is not None:
            return OK(self.link)
        link = self.core.getLink()
        return link.connectEngine(self).map(self.__cacheLink)

    def __cacheCurrentEnv(self, lr):
        env = lr.stdout.strip()
        self.logAdapter.debug("Current environment: '%s'" % (env))
        self.currentEnv = env
        return lr

    def __cacheLink(self, link):
        self.link = link
        return link

    def __start(self, *args):
        return self.getDriver().startMachine(self.getDriverID()).map(self.chainSelf)

    def __suspend(self, *args):
        return self.getDriver().suspendMachine(self.getDriverID()).map(self.chainSelf)

    def __terminate(self, *args):
        return self.getDriver().terminateMachine(self.getDriverID()).map(self.chainSelf)

    def __halt(self, *args):
        return self.getDriver().haltMachine(self.getDriverID()).map(self.chainSelf)

    def __delete(self, *args):
        return self.getDriver().deleteMachine(self.getDriverID()).map(self.chainSelf)

    def __readDriverNetworkInfo(self):
        return self.getDriver().readMachineNetworkInfo(self.getDriverID())

    def __configure(self):
        return self.getDriver().configureMachine(self.getDriverID(), self)

    def shell(self):
        return self.readLink() >> Link.interactive

    def __waitForNetwork(self):
        self.logAdapter.info("Waiting for machine network...")
        return self.getDriver().readMachineWaitForNetwork(self.getDriverID())

    def __waitForReady(self):
        self.logAdapter.info("Waiting for machine to boot...")
        return self.readLink()

    def setHostname(self):
        self.logAdapter.info("Configuring machine hostname")
        dns = self.getDNSName()
        echoCmd = "echo %s > /etc/hostname" % dns
        hostCmd = "hostname %s" % dns
        hostsCmd = "sed -i \"s/127.0.1.1.*/127.0.1.1\t%s/g\" /etc/hosts" % self.name
        # hostsCmd = "hostnamectl set-hostname %s" % dns
        serviceCmd = "service hostname restart"
        cmd = "sudo -- /bin/sh -c '%s && %s && %s && %s'" % (
            echoCmd, hostsCmd, hostCmd, serviceCmd)
        cmds = list(map(defer(self.link.runCommand, stream=True, sudo=False), [cmd]))
        return Try.sequence(cmds)

    def exposePort(self, local_port, public_port, scheme):
        keyfile = self.core.getInsecureKeyFile()
        keyfile = inner(keyfile)
        forward_descr = "0.0.0.0:%s:127.0.0.1:%s" % (public_port, local_port)
        engineIP = self.getSSHIP()
        enginePort = str(self.getSSHPort())
        cmdPath = "ssh"
        cmdArgs = ["ssh", "-N", "-L", forward_descr, engineIP, "-l", "substance", "-p", enginePort,
                   "-o", "StrictHostKeyChecking=no", "-o", "UserKnownHostsFile=/dev/null", "-i", keyfile]
        sudo = False
        if public_port < 1024 and not isWithinWindowsSubsystem():
            sudo = True
        self.logAdapter.info(
            "Exposing port %s as %s; kill the process (CTRL-C) to un-expose.", local_port, public_port)
        with closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
            s.connect(("8.8.8.8", 80))
            ip = s.getsockname()[0]
            if scheme:
                self.logAdapter.info(
                    "%s://%s:%s is now available.", scheme, ip, public_port)
            else:
                self.logAdapter.info(
                    "Others can now connect to port %s via IP %s.", public_port, ip)
        Shell.execvp(cmdPath, cmdArgs, {}, sudo=sudo)

    def getOrchestrator(self):
        if self.subenvConfig:
            return self._getOrchestrator()

        return self.envLoadCurrent() \
            .then(self.loadSubenvConfigFile) \
            .catch(lambda err: OK(None)) \
            .then(self._getOrchestrator)

    def _getOrchestrator(self):
        orchestrator = self.subenvConfig.get('orchestrator') if self.subenvConfig else None
        logger.debug("Orchestrator: %s" % orchestrator)
        if orchestrator and orchestrator == Orchestrators.COMPOSE:
            return OK(Compose(self))
        elif orchestrator and orchestrator != Orchestrators.DOCKWRKR:
            return Fail(InvalidOrchestrator('%s is not a known orchestrator' % orchestrator))
        else:
            return OK(Dockwrkr(self))

    def getOrchestratorCommand(self, command, *args):
        orch = self.getOrchestrator()
        if orch.isFail():
            raise orch.value
        else:
            orch = orch.getOK()

        if orch:
            return getattr(orch, command)(*args)
        else:
            return ''

    def orchestrate(self, command, *args):
        commands = self.getOrchestratorCommand(command, *args)
        if len(commands) > 0:
            return list(map(lambda x: "(subenv run %s)" % x, commands))

    def envSwitch(self, subenvName, restart=False):
        self.logAdapter.info(
            "Switch engine '%s' to subenv '%s'" % (self.name, subenvName))

        self.currentEnv = subenvName
        self.loadSubenvConfigFile()

        cmds = CommandList([
            "subenv init '/substance/devroot/%s'" % (subenvName),
            "subenv use '%s'" % (subenvName),
        ])

        if restart:
            cmds.append(self.orchestrate('resetAll'))
            cmds.append(self.orchestrate('login'))
            cmds.append(self.orchestrate('startAll'))

        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False) \
            .then(self.envRegister)

    def envRegister(self):
        return self.readLink() \
            .bind(Link.runCommand, 'subenv vars', stream=False, capture=True) \
            .bind(self.__envRegister)

    def __envRegister(self, lr):
        vars = lr.stdout.split("\n")
        env = Try.attempt(parseDotEnv, vars)
        if env.isFail():
            return env
        env = env.getOK()
        if 'SUBENV_FQDN' in env:
            return SubHosts.register(env['SUBENV_FQDN'], self.getPublicIP())
        elif 'SUBENV_NAME' in env:
            return SubHosts.register(env['SUBENV_NAME'] + self.core.config.get('tld'), self.getPublicIP())

    def envLoadCurrent(self):
        cmds = CommandList(["subenv current"])
        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=False, sudo=False, capture=True) \
            .map(self.__cacheCurrentEnv) \
            .catch(lambda err: Fail(EnvNotDefinedError("No current subenv is set. Check 'switch' for detais."))) \
            .then(lambda: self)

    def envStart(self, reset=False, containers=[]):
        cmds = CommandList()

        if reset:
            cmds.append(self.getOrchestratorCommand('resetAll'))

        if len(containers) > 0:
            self.logAdapter.info("Starting %s container(s)" %
                                 (' '.join(containers)))
            cmds.append(self.orchestrate('start', containers))
        else:
            self.logAdapter.info("Starting all containers...")
            cmds.append(self.orchestrate('startAll'))

        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)

    def envRestart(self, time=10, containers=[]):
        cmds = CommandList()

        if len(containers) > 0:
            self.logAdapter.info("Restarting %s containers" %
                                 (' '.join(containers)))
            cmds.append(self.orchestrate('restart', containers, time))
        else:
            self.logAdapter.info("Restarting all containers...")
            cmds.append(self.orchestrate('restartAll', time))

        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)

    def envRecreate(self, time=10, containers=[]):
        cmds = CommandList()

        if len(containers) > 0:
            self.logAdapter.info("Recreating %s containers" %
                                 (' '.join(containers)))
            cmds.append(self.orchestrate('recreate', containers, time))
        else:
            self.logAdapter.info("Recreating all containers...")
            cmds.append(self.orchestrate('recreateAll', time))

        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)

    def envShell(self, container=None, user=None, cwd=None):
        if container:
            return self.envEnter(container, user, cwd)
        else:
            return self.readLink().bind(Link.interactive)

    def envEnter(self, container, user=None, cwd=None):
        self.logAdapter.info("Entering %s container..." % (container))
        return self.envExec(container, ["exec /bin/bash"], cwd, user)

    def envExec(self, container, cmd, cwd=None, user=None):
        cmd = CommandList(self.orchestrate('exec', container, cmd, cwd, user))
        return self.readLink().bind(Link.runCommand, cmd=cmd.logicAnd(), interactive=True, stream=True, shell=False, capture=False)

    def envRun(self, task, args):
        cmd = CommandList(self.orchestrate('run', task, args))
        return self.readLink().bind(Link.runCommand, cmd=cmd.logicAnd(), interactive=True, stream=True, shell=False, capture=False)

    def envExecAlias(self, alias, args):
        aliases = self.config.get('aliases')
        if self.subenvConfig.get('aliases'):
            aliases = mergeDictOverwrite(self.config.get('aliases'),
                                         self.subenvConfig.get('aliases'))

        if aliases and alias in aliases:
            cmd = aliases[alias]
            args = cmd['args'] + args
            return self.envExec(container=cmd['container'], cmd=args, cwd=cmd['cwd'], user=cmd['user'])
        return Fail(InvalidCommandError("Invalid command '%s' specified for '%s'.\n\nUse 'substance help' for available commands." % (alias, 'substance')))

    def envStop(self, time=10, containers=[]):
        cmds = CommandList()

        if len(containers) > 0:
            self.logAdapter.info("Stopping %s container(s)" %
                                 (' '.join(containers)))

            cmds.append(self.orchestrate('stop', containers, time))
        else:
            self.logAdapter.info("Stopping containers...")
            cmds.append(self.orchestrate('stopAll', time))

        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=True, sudo=False)

    def envStatus(self, full=False):
        if full:
            cmds = CommandList([self.orchestrate('status')])
            return self.readLink() \
                .bind(Link.runCommand, cmds.logicAnd(), stream=False, sudo=False, capture=True) \
                .map(self.__envStatus)
        else:
            return OK(self.__envStatus())

    def envCleanup(self):
        cmds = CommandList()
        cmds.append('docker system prune -a')
        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), interactive=True, stream=True, shell=False, capture=False)

    def envLogs(self, parts=[], pattern=None, follow=True, lines=None):

        if not pattern:
            pattern = "%s*.log" % ('-'.join(parts))

        cmds = CommandList()
        cmd = "tail"
        if follow:
            cmd += " -f"
        if lines:
            cmd += " -n %s" % int(lines)

        cmd += " \"logs/%s\"" % pattern

        cmds.append('subenv run %s' % cmd)
        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=True, capture=False, sudo=False)

    def envListLogs(self, parts=[], pattern=None):

        if not pattern:
            pattern = "%s*.log" % ('-'.join(parts))

        cmds = CommandList()
        cmd = "ls -1 \"logs/%s\" | xargs -n1 basename" % pattern

        cmds.append('subenv run %s' % cmd)
        return self.readLink() \
            .bind(Link.runCommand, cmds.logicAnd(), stream=True, capture=False, sudo=False)

    def __envStatus(self, containers=None):
        return {
            'engine': self,
            'containers': containers.stdout + containers.stderr if containers else None
        }

    def envDocker(self, command):
        cmd = "docker %s" % command
        logger.debug("DOCKER: %s" % cmd)
        return self.readLink().bind(Link.runCommand, cmd, stream=True, interactive=True, shell=False)

    def getEngineFolders(self):
        # XXX Dynamic mounts / remove hardcoded values.
        devroot = self.config.get('devroot')
        pfolder = EngineFolder(
            name='devroot',
            mode=devroot.get('mode'),
            hostPath=os.path.expanduser(devroot.get('path')),
            guestPath='/substance/devroot',
            uid=1000,
            gid=1000,
            umask="0022",
            excludes=devroot.get('excludes', []),
            syncArgs=devroot.get('syncArgs', [])
        )
        return [pfolder]

    def mountFolders(self):
        self.logAdapter.info("Mounting engine folders")
        folders = self.getEngineFolders()
        return Try.of(list(map(self.mountFolder, folders)))

    def mountFolder(self, folder):
        mountCmd = "mount -t vboxsf -o umask=%(umask)s,gid=%(gid)s,uid=%(uid)s %(name)s %(guestPath)s" % folder.__dict__
        mkdirCmd = "mkdir -p %(guestPath)s && chown -R %(uid)s:%(gid)s %(guestPath)s" % folder.__dict__
        # XXX Make this non VBOX specific
        chain = self.link.runCommand(mkdirCmd, sudo=True)
        if folder.mode is 'sharedfolder':
            chain = chain.then(defer(self.link.runCommand, mountCmd, sudo=True))
        return chain

    chainSelf = chainSelf

    def __repr__(self):
        return "Engine(%s)" % self.name



import unittest
import os
import tempfile
import tests
import random
import string
from substance.monads import *
from substance import (Core, EngineProfile, Engine, Shell)
from substance.utils import writeToFile, parseDotEnv, readDotEnv

from substance.subenv.constants import (SPECDIR, ENVFILE, CODELINK)
from substance.subenv import SubenvAPI


class TestSubenv(tests.TestBase):

    def setUp(self):
        self.basePath = self.addTemporaryDir()
        self.projectsBasePath = self.addTemporaryDir()
        self.subenvBasePath = self.addTemporaryDir()

    def tearDown(self):
        pass
        if self.basePath:
            Shell.nukeDirectory(self.basePath).catch(tests.TestBase.raiser)
        if self.subenvBasePath:
            Shell.nukeDirectory(self.subenvBasePath).catch(tests.TestBase.raiser)
        if self.projectsBasePath:
            Shell.nukeDirectory(self.projectsBasePath).catch(tests.TestBase.raiser)

    def setUpDefaultCore(self):
        self.core = Core(basePath=self.basePath)
        self.assertIsInstance(self.core.initialize(), OK)
        self.api = SubenvAPI(self.subenvBasePath)
        self.assertIsInstance(self.api.initialize(), OK)
        return self.core

    def makeEnv(self):

        name = self.randString()

        ppath = os.path.join(self.projectsBasePath, name)
        specpath = os.path.join(ppath, SPECDIR)

        op = Shell.makeDirectory(ppath) \
            .then(defer(Shell.makeDirectory, specpath))

        self.assertIsInstance(op, OK)

        dirs = ['conf', os.path.join('conf', 'web'), os.path.join('conf', 'cron'), os.path.join('conf', 'logrotate')]
        ops = [Try.attempt(Shell.makeDirectory, os.path.join(specpath, x)) for x in dirs]

        self.assertIsInstance(Try.sequence(ops), OK)

        tpl1 = "{{SUBENV_ENVPATH}}\n{{SUBENV_BASEPATH}}"
        tpl2 = 'This is a template with no vars.'

        customVar = self.randString()

        envTpl = 'name="%(name)s"\ncustom="%(custom)s"'

        filedata = {}
        filedata[ENVFILE] = envTpl % {'name': name, 'custom': customVar}
        filedata["tpl1.jinja"] = tpl1
        filedata[os.path.join("conf", "tpl2.jinja")] = tpl2

        ops = []
        for file, data in filedata.items():
            ops.append(Try.attempt(writeToFile, os.path.join(specpath, file), data.encode()))

        self.assertIsInstance(Try.sequence(ops), OK)

        return {'name': name, 'custom': customVar, 'path': ppath, 'files': filedata, 'dirs': dirs}

    def testInit(self):
        self.setUpDefaultCore()
        self.env = self.makeEnv()

        custom2 = self.randString()
        op = self.api.init(self.env['path'], {'custom2': custom2})
        self.assertIsInstance(op, OK)

        envPath = os.path.join(self.subenvBasePath, "envs", self.env['name'])

        self.assertTrue(os.path.isdir(envPath))

        for f, fd in self.env['files'].items():
            root, ext = os.path.splitext(f)
            if ext == '.jinja':
                self.assertTrue(os.path.isfile(os.path.join(envPath, root)))
            elif f == ENVFILE:
                self.assertTrue(os.path.isfile(os.path.join(envPath, f)))
                data = readDotEnv(os.path.join(envPath, f))
                self.assertEqual(data['name'], self.env['name'])
                self.assertEqual(data['custom'], self.env['custom'])
                self.assertEqual(data['custom2'], custom2)
                self.assertEqual(data['SUBENV_NAME'], self.env['name'])
                self.assertEqual(data['SUBENV_BASEPATH'], os.path.join(self.projectsBasePath, self.env['name']))
                self.assertEqual(data['SUBENV_SPECPATH'], os.path.join(self.projectsBasePath, self.env['name'], SPECDIR))
            else:
                self.assertTrue(os.path.isfile(os.path.join(envPath, f)))
                with open(os.path.join(envPath, f), 'r') as fh:
                    data = fh.read()
                self.assertEqual(fd, data)

        for d in self.env['dirs']:
            self.assertTrue(os.path.isdir(os.path.join(envPath, d)))

        ls = self.api.ls()
        self.assertIsInstance(ls, OK)
        ls = ls.getOK()
        names = [x.name for x in ls]
        self.assertIn(self.env['name'], names)

        op = self.api.delete(self.env['name'])
        self.assertIsInstance(op, OK)

        ls = self.api.ls()
        self.assertIsInstance(ls, OK)
        ls = ls.getOK()
        names = [x.name for x in ls]
        self.assertNotIn(self.env['name'], names)

    def randString(self):
        return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))



import sys
import os
import logging
from collections import OrderedDict

from substance.monads import *
from substance.logs import *
from substance.shell import Shell
from substance.engine import Engine
from substance.link import Link
from substance.box import Box
from substance.db import DB
from substance.constants import Tables, EngineStates, DefaultEngineBox
from substance.utils import (
    readYAML,
    writeYAML,
    readSupportFile,
    getSupportFile,
    streamDownload,
    makeXHRRequest,
    sha1sum
)
from substance.path import (getHomeDirectory)
from substance.config import (Config)
from substance.driver.virtualbox import VirtualBoxDriver
from substance.exceptions import (
    FileSystemError,
    FileDoesNotExist,
    EngineNotFoundError,
    EngineExistsError,
    EngineNotRunning
)
import requests

logger = logging.getLogger(__name__)

class Core(object):

    def __init__(self, configFile=None, basePath=None):
        self.basePath = os.path.abspath(basePath) if basePath else getHomeDirectory('.substance')
        self.enginesPath = os.path.join(self.basePath, "engines")
        self.boxesPath = os.path.join(self.basePath, "boxes")
        self.dbFile = os.path.join(self.basePath, "db.json")

        configFile = configFile if configFile else "substance.yml"
        configFile = os.path.join(self.basePath, configFile)
        self.config = Config(configFile)

        self.insecureKey = None
        self.insecurePubKey = None

        self.assumeYes = False
        self.initialized = False

    def getBasePath(self):
        return self.basePath

    def getEnginesPath(self):
        return self.enginesPath

    def getBoxesPath(self):
        return self.boxesPath

    def getDbFile(self):
        return self.dbFile

    def initialize(self):
        if self.initialized:
            return OK(None)
        return self.assertPaths().then(self.assertConfig).then(self.initializeDB).then(defer(self.setInitialized, b=True))

    def setInitialized(self, b):
        self.initialized = b

    def assertPaths(self):
        return OK([self.basePath, self.enginesPath, self.boxesPath]).mapM(Shell.makeDirectory)

    def assertConfig(self):
        return self.config.loadConfigFile()  \
            .catchError(FileDoesNotExist, self.makeDefaultConfig)

    def getDefaultConfig(self):
        defaults = OrderedDict()
        defaults['assumeYes'] = False
        defaults['drivers'] = ['virtualbox']
        defaults['tld'] = '.dev'
        defaults['devroot'] = getHomeDirectory('substance')
        defaults['current'] = OrderedDict()
        defaults['engine'] = None
        defaults['subenv'] = None
        return defaults

    def makeDefaultConfig(self, data=None):
        logger.info("Generating default substance configuration in %s",
                    self.config.getConfigFile())
        defaults = self.getDefaultConfig()
        for kkk, vvv in defaults.items():
            self.config.set(kkk, vvv)
        self.config.set("basePath", self.basePath)
        return self.config.saveConfig()

    # -- Use

    def setUse(self, engine, subenvName=None):
        ops = [self.setCurrentEngine(engine)]
        if subenvName:
            ops.append(engine.envSwitch(subenvName))
        return Try.sequence(ops)

    def setCurrentEngine(self, engine):
        current = self.config.get('current')
        current.update({'engine': engine.name})
        self.config.set('current', current)
        return OK(self)

    def readCurrentEngineName(self):
        current = self.config.get('current', {})
        name = current.get('engine', None)
        if not name:
            return Fail(EngineNotFoundError("No current engine is specified. Check the 'use' command for details."))
        return OK(name)

    def loadCurrentEngine(self, name=None):
        current = self.config.get('current', {})
        engineName = name
        if not engineName:
            engineName = current.get('engine', None)

        if not engineName:
            return Fail(EngineNotFoundError("No current engine is specified. Check the 'use' command for details."))

        engine = self.loadEngine(engineName) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.loadState)
        if engine.isFail():
            return engine

        engine = engine.getOK()

        if engine.state is not EngineStates.RUNNING:
            return Fail(EngineNotRunning("Engine '%s' is not running." % engine.name))

        return OK(engine)

    # -- Runtime

    def setAssumeYes(self, ay):
        self.assumeYes = True
        return True

    def getAssumeYes(self):
        if self.config.get('assumeYes', False):
            return True
        elif self.assumeYes:
            return True
        return False

    def getDefaultBoxString(self):
        return DefaultEngineBox

    # -- Engine library management

    def getEngines(self):
        ddebug("getEngines()")
        dirs = [d for d in os.listdir(self.enginesPath) if os.path.isdir(
            os.path.join(self.enginesPath, d))]
        return OK(dirs)

    def loadEngines(self, engines=[]):
        return OK([self.loadEngine(x) for x in engines])

    def loadEngine(self, name):
        enginePath = os.path.join(self.enginesPath, name)
        if not os.path.isdir(enginePath):
            return Fail(EngineNotFoundError("Engine \"%s\" does not exist." % name))
        else:
            return OK(Engine(name, enginePath=enginePath, core=self))

    def createEngine(self, name, config=None, profile=None):
        enginePath = os.path.join(self.enginesPath, name)
        newEngine = Engine(name, enginePath=enginePath, core=self)
        return newEngine.create(config=config, profile=profile)

    def removeEngine(self, name):
        return self.loadEngine(name) \
            >> Engine.remove

    # -- Driver handling

    def getDrivers(self):
        return self.config.get('drivers', [])

    def validateDriver(self, driver):
        if driver in self.getDrivers():
            return OK(driver)
        return Fail(ValueError("Driver '%s' is not a valid driver."))

    def getDriver(self, name):
        cls = {
            'virtualbox': VirtualBoxDriver
        }.get(name, 'virtualbox')
        driver = cls(core=self)
        return driver

    # -- Link handling

    def getLink(self, type="ssh"):
        link = Link(keyFile=self.getInsecureKeyFile(), keyFormat='RSA')
        return link

    # -- Database

    def getDB(self):
        return self.db

    def initializeDB(self):
        db = DB(self.dbFile)
        db = db.initialize()
        if db.isFail():
            return db
        self.db = db.getOK()
        return OK(self.db)

    # -- Box handling

    def readBox(self, boxstring):
        return Box.parseBoxString(boxstring) \
            .map(lambda p: Box(core=self, **p))

    def pullBox(self, box):
        return box.fetch()

    def removeBox(self, box):
        return box.delete()

    def getBoxes(self):
        return self.getDB().getBoxRecords() \
            .mapM(lambda r: OK(Box(self, r.get('name'), r.get('version'), r.get('namespace'), r.get('registry'), r.get('boxstring'), r.get('archiveSHA1'))))

    def getInsecureKeyFile(self):
        return getSupportFile('support/substance_insecure')

    def getInsecurePubKeyFile(self):
        return getSupportFile('support/substance_insecure.pub')




class Constants(object):

    class ConstError(TypeError):
        pass

    def __init__(self, **kwargs):
        for name, value in list(kwargs.items()):
            super(Constants, self).__setattr__(name, value)

    def __setattr__(self, name, value):
        if name in self.__dist__:
            raise self.ConstError("Can't rebind const(%s)" % name)
        self.__dict__[name] = value

    def __delattr__(self, name):
        if name in self.__dict__:
            raise self.ConstError("Can't unbind const(%s)" % name)
        raise NameError(name)


Tables = Constants(
    BOXES="boxes"
)

DefaultEngineBox = 'turbulent/substance-box:1.0'

EngineStates = Constants(
    RUNNING="running",
    STOPPED="stopped",
    SUSPENDED="suspended",
    UNKNOWN="unknown",
    INEXISTENT="inexistent"
)

Syncher = Constants(
    UP=">>",
    DOWN="<<",
    BOTH="<>"
)

Orchestrators = Constants(
    DOCKWRKR="dockwrkr",
    COMPOSE="docker-compose"
)



import sys
import yaml
import requests
import tempfile
import shutil
import logging
import collections
import functools
import os
import errno
import hashlib
import tarfile
from math import floor
from time import time
from pkg_resources import Requirement, resource_filename, require
from substance.exceptions import (
    ConfigSyntaxError,
    FileSystemError,
    FileDoesNotExist
)

logger = logging.getLogger(__name__)

_yaml_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


def _dict_representer(dumper, data):
    return dumper.represent_dict(iter(data.items()))


def _dict_constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))


yaml.add_representer(collections.OrderedDict, _dict_representer)
yaml.add_constructor(_yaml_mapping_tag, _dict_constructor)


def writeToFile(filename, the_bytes):
    try:
        with open(filename, "wb") as fh:
            fh.write(the_bytes)
    except Exception as err:
        raise FileSystemError("Failed to write to %s : %s" % (filename, err))


def writeYAML(filename, data):
    try:
        with open(filename, "w") as fileh:
            fileh.write(yaml.dump(data, default_flow_style=False))
    except Exception as err:
        raise FileSystemError("Failed to write to %s : %s" % (filename, err))


def readYAML(filename):
    try:
        with open(filename, "r") as stream:
            contents = yaml.load(stream)
            return contents
    except yaml.YAMLError as exc:
        msg = "Syntax error in file %s"
        if hasattr(exc, 'problem_mark'):
            mark = exc.problem_mark
            msg += " Error position: (%s:%s)" % (mark.line +
                                                 1, mark.column + 1)
        raise ConfigSyntaxError(msg)
    except Exception as err:
        raise FileSystemError(
            "Failed to read configuration file %s : %s" % (filename, err))


def getSupportFile(filename):
    return resource_filename(Requirement.parse('substance'), filename)


def getPackageVersion():
    return require('substance')[0].version


def readSupportFile(filename):
    try:
        keyfile = getSupportFile(filename)
        with open(keyfile, "r") as fileh:
            keydata = fileh.read()
            return keydata
    except IOError as err:
        raise FileDoesNotExist("File not found: %s" % keyfile)
    except Exception as err:
        raise FileSystemError("Failed to read: %s" % keyfile)


def streamDownload(url, destination):
    basename = os.path.basename(destination)

    try:
        with open(destination, 'wb') as fd:
            r = requests.get(url, stream=True, timeout=5)
            r.raise_for_status()
            contentLength = int(r.headers.get('content-length'))

            if contentLength is None:
                fd.write(r.content)
            else:
                progress = 0
                startTime = time()
                chunkSize = floor(contentLength / 100)
                for chunk in r.iter_content(chunk_size=chunkSize):
                    if chunk:
                        progress += len(chunk)
                        fd.write(chunk)
                        done = int(50 * progress / contentLength)
                        elapsed = (time() - startTime)
                        speed = humanReadableBytes(progress / elapsed)
                        left = humanReadableBytes(contentLength - progress)
                        sys.stdout.write(
                            "\r%s [%s%s] %s/s %s left" % (basename,  '=' * done, ' ' * (50 - done), speed, left))
                        sys.stdout.flush()
                sys.stdout.write("\n")
            fd.flush()
    except (Exception, KeyboardInterrupt) as err:
        if os.path.exists(destination):
            os.remove(destination)
        raise(err)
    return destination


def humanReadableBytes(num):
    for x in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0:
            return "%3.1f%s" % (num, x)
        num /= 1024.0
    return "%3.1f%s" % (num, 'TB')


def sha1sum(filename):
    sha = hashlib.sha1()
    bufsize = 65536
    with open(filename, 'rb') as fd:
        while True:
            data = fd.read(bufsize)
            if not data:
                break
            sha.update(data)
    return sha.hexdigest()


def makeHttpRequest(url, params=None):
    r = requests.get(url, params=params)
    r.raise_for_status()
    return r.content


def makeXHRRequest(url, params=None):
    r = requests.get(url, params=params)
    r.raise_for_status()
    payload = r.json()
    return payload


def untar(filename, destination=None):
    tar = tarfile.open(filename)
    tar.extractall(destination)
    tar.close()
    return filename


def mergeDict(a, b, path=None):
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                mergeDict(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


# merge 2 dicts and overwrite target keys with source keys if collision
def mergeDictOverwrite(target, source):
    merged = target.copy()
    merged.update(source)
    return merged


def readDotEnv(filepath, env={}):
    with open(filepath) as f:
        return parseDotEnv(f, env)


def parseDotEnv(dotenvdata, env={}):
    for line in dotenvdata:
        line = line.strip()
        if not line or line.startswith('#') or '=' not in line:
            continue
        k, v = line.split('=', 1)
        v = v.strip("'").strip('"')
        env[k] = v
    return env


def makeSymlink(source, link, force=False):
    try:
        os.symlink(source, link)
    except OSError as e:
        if e.errno == errno.EEXIST and force:
            os.remove(link)
            os.symlink(source, link)
        else:
            raise e


def readSymlink(link):
    return os.readlink(link)


def flatten(l):
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


class CommandList(list):
    def commands(self):
        cmds = list(filter(lambda x: x is not None, self))
        return flatten(cmds)

    def joined(self, by=' '):
        return by.join(self.commands())

    def logicAnd(self):
        return self.joined(' && ')

    def logicOr(self):
        return self.joined(' || ')


class Memoized(object):
    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''
    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)


class DictCompat(dict):
    def iteritems(self):
        return self.items()



import logging

from substance.monads import *
from substance.logs import *
from substance.constants import Tables
from tinydb import (
    TinyDB,
    Query,
    where
)

logger = logging.getLogger(__name__)

class DB(object):

    def __init__(self, dbFile):
        self.dbFile = dbFile

    def initialize(self):
        db = Try.attempt(lambda: TinyDB(self.dbFile))
        if db.isFail():
            return db
        self.db = db.getOK()
        return OK(self)

    # -- Box methods

    def updateBoxRecord(self, box, record):
        return self.getBoxRecord(box) \
            .bind(self._updateBoxRecord, record=record)

    def removeBoxRecord(self, box):
        logger.debug("Removing box record: %s" % (box))
        return self.getBoxRecord(box).bind(lambda r: self.remove(Tables.BOXES, eids=[r.eid]) if r else OK(None))

    def _updateBoxRecord(self, element, record):
        if element:
            logger.debug("Updating box record: %s : %s" % (element, record))
            return self.update(Tables.BOXES, record, eids=[element.eid])
        else:
            logger.info("Inserting box record: %s : %s" % (element, record))
            return self.insert(Tables.BOXES, record)

    def getBoxRecord(self, box):
        q = Query()
        q = (q.name == box.name) & (q.version == box.version) & (
            q.namespace == box.namespace)
        ret = self.get(Tables.BOXES, q)
        return ret

    def getBoxRecords(self):
        return self.all(Tables.BOXES)

    # -- DB mapping

    def tables(self):
        return Try.attempt(lambda: self.db.tables())

    def purge_tables(self):
        return Try.attempt(lambda: self.db.purge_tables())

    def all(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('all', *args, **kwargs)

    def search(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('search', *args, **kwargs)

    def get(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('get', *args, **kwargs)

    def remove(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('remove', *args, **kwargs)

    def contains(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('contains', *args, **kwargs)

    def update(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('update', *args, **kwargs)

    def insert(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('insert', *args, **kwargs)

    def insert_multiple(self, table, *args, **kwargs):
        return self.readTable(table) >> self.proxyTableCommand('insert_multiple', *args, **kwargs)

    def proxyTableCommand(self, command, *args, **kwargs):
        def proxy(table):
            f = getattr(table, command)
            return Try.attempt(lambda: f(*args, **kwargs))
        return proxy

    def readTable(self, table):
        return OK(self.db.table(table))



import os
import time
import platform
import logging
import queue
import fnmatch
from collections import OrderedDict

from substance.monads import *
from substance.logs import *
from substance.shell import Shell
from substance.exceptions import (SubstanceError)
from substance.utils import getSupportFile
from substance.path import pathComponents
from substance.constants import Syncher
from substance.syncher import BaseSyncher

from subwatch.watcher import (LocalWatchAgent, RemoteWatchAgent)
from subwatch.events import WatchEventHandler
from subwatch.events import (WatchEvent, EVENT_TYPE_MODIFIED)

from tornado import ioloop
from functools import reduce

logger = logging.getLogger(__name__)


class SubwatchSyncher(BaseSyncher):

    PARTIAL_DIR = '.~subsync~'

    def __init__(self, engine, keyfile):
        super(SubwatchSyncher, self).__init__(engine, keyfile)
        self.localAgent = LocalWatchAgent(self.processLocalEvent)
        self.remoteAgent = RemoteWatchAgent(self.processRemoteEvent)
        self.synching = {Syncher.UP: {}, Syncher.DOWN: {}}
        self.toSync = {Syncher.UP: {}, Syncher.DOWN: {}}
        self.syncPeriod = 0.3
        self.schedule = {Syncher.UP: False, Syncher.DOWN: False}
        self.excludes = []

    def getFolders(self):
        return reduce(lambda acc, x: acc + [x] if x.mode == 'rsync' else acc, self.engine.getEngineFolders(), [])

    def getFolderFromPath(self, path, direction=None):
        direction = Syncher.UP if not direction else direction

        for folder in self.getFolders():
            if direction is Syncher.UP and folder.hostPath == path:
                return OK(folder)
            elif direction is Syncher.DOWN and folder.guestPath == path:
                return OK(folder)
        return Fail(NameError("%s is not a folder." % path))

    def start(self, direction=Syncher.BOTH, path=''):
        op = self.ensureKeyPerms() \
            .then(defer(self.initialSync, direction=direction))
        if op.isFail():
            return op

        folders = self.getFolders()

        if len(folders) == 0:
            return Fail(SubstanceError("No rsync folders definedfor synchronization in engine configuration."))

        logger.info("Starting substance syncher (%s)" % direction)

        for folder in folders:
            if direction in [Syncher.UP, Syncher.BOTH]:
                self.localAgent.startWatching(folder.hostPath)
            if direction in [Syncher.DOWN, Syncher.BOTH]:
                self.remoteAgent.startWatching(folder.guestPath)

        self.remoteAgent.connect("ws://%s:%s" %
                                 (self.engine.getPublicIP(), 1500))

        try:
            self.remoteAgent.start()
            ioloop.IOLoop.current().start()
            return OK(None)
        except KeyboardInterrupt:
            return self.stop()
        except Exception as err:
            logger.error(err)
            return Fail(err)
        finally:
            return self.stop()

        return OK(None)

    def stop(self):
        self.localAgent.stop()
        self.remoteAgent.stop()
        ioloop.IOLoop.current().stop()
        return OK(None)

    def scheduleSync(self, direction):
        if self.schedule[direction] is False:
            self.schedule[direction] = True
            ioloop.IOLoop.current().call_later(self.syncPeriod, defer(
                self.incrementalSync, direction=direction))

    def processLocalEvent(self, event):
        return self.expungeSynching(Syncher.UP).then(defer(self.processEvent, event, Syncher.UP))

    def processRemoteEvent(self, event):
        return self.expungeSynching(Syncher.DOWN).then(defer(self.processEvent, event, Syncher.DOWN))

    def expungeSynching(self, direction):
        dirs = self.synching[direction]
        for dir in list(dirs.keys()):
            t = self.synching[direction][dir]
            if (time.time() - t) > 0:
                logger.debug("Expiring ignore of (%s)%s" % (direction, dir))
                del self.synching[direction][dir]
        return OK(None)

    def ignoreSync(self, direction, folder, paths=[], timeout=1):
        for path in paths:
            logger.debug("Ignoring (%s)%s for %ss" %
                         (direction, path, timeout))
            self.synching[direction][path] = time.time() + timeout

    def processEvent(self, event, direction=None):
        direction = Syncher.UP if direction is None else direction

        path = event.getRelativePath()
        folder = self.getFolderFromPath(event.watchPath, direction)

        if folder.isFail():
            logger.error("%s" % folder)
            return folder
        else:
            folder = folder.getOK()

        if event.type == EVENT_TYPE_MODIFIED and event.isDirectory:
            return

        if self.fileMatch(path, self.getExcludes()):
            logger.info("Ignored (excluded) event (%s)%s" % (direction, path))
            return

        if os.path.dirname(path) in self.synching[direction]:
            logger.debug("Ignored (synching) event of (%s)%s" %
                         (direction, path))
            return

        if folder not in self.toSync[direction]:
            self.toSync[direction][folder] = []

        if not event.isDirectory:
            path = os.path.dirname(path)

        if path not in self.toSync[direction][folder]:
            if path not in self.synching[direction]:
                self.ignoreSync(self.inverse(direction), folder, [path])
            self.toSync[direction][folder].append(path)
            self.scheduleSync(direction)

    def inverse(self, direction):
        if direction == Syncher.UP:
            return Syncher.DOWN
        else:
            return Syncher.UP

    def fileMatch(self, filename, excludes=[]):
        for ex in excludes:
            if fnmatch.fnmatch(filename, ex):
                return True
        return False

    def initialSync(self, direction):
        folders = self.getFolders()
        chain = OK(None)
        for folder in folders:
            if direction in [Syncher.UP, Syncher.BOTH]:
                chain = chain.then(
                    defer(self.syncFolder, folder=folder, direction=Syncher.UP, paths=['/']))
            if direction in [Syncher.DOWN, Syncher.BOTH]:
                chain = chain.then(
                    defer(self.syncFolder, folder=folder, direction=Syncher.DOWN, paths=['/']))
        return chain

    def incrementalSync(self, direction):
        folders = self.toSync[direction].copy()
        seq = [self.syncFolder(
            folder, direction, folders[folder], incremental=True) for folder in folders]
        self.schedule[direction] = False
        return Try.sequence(seq)

    def syncFolder(self, folder, direction, paths, incremental=False):
        filters = "\n".join(self.makeFilters(folder, paths))

        if paths:
            logger.info("Synchronizing %s %s %s:%s" % (
                folder.hostPath, direction, self.engine.name, folder.guestPath))

        self.toSync[direction].pop(folder, None)

        syncher = Rsync()
        syncher.setTransport(self.engine.getSSHPort(), self.keyfile)
        syncher.setFilters(self.makeFilters(folder, paths))
        syncher.setLongOption('inplace', True)
        syncher.setLongOption('temp-dir', '/tmp')
        syncher.setLongOption('itemize-changes', True)
        syncher.setLongOption('log-format', '%o\t%n %L')

        if incremental is True:
            syncher.setLongOption('delete', True)

        if direction == Syncher.UP:
            syncher.setFrom(folder.hostPath)
            syncher.setTo(folder.guestPath,
                          self.engine.getSSHIP(), "substance")
        elif direction == Syncher.DOWN:
            syncher.setFrom(folder.guestPath,
                            self.engine.getSSHIP(), "substance")
            syncher.setTo(folder.hostPath)
        else:
            return Fail(ValueError("Sync direction must be 'up' or 'down'"))

        return syncher.sync() \
            .then(dinfo("Finished sync of %s %s %s:%s" % (folder.hostPath, direction, self.engine.name, folder.guestPath))) \
            .then(lambda: OK(folder))

    def getExcludes(self):
        exs = self.excludes
        plus = []
        plus.append('*.~subsync~')
        plus.append('*.~subsync~/**')
        plus.append('*.*.swo')
        plus.append('*.*.swp')
        plus.append('*.*.swpx')
        plus.append('*.*.ssh')
        plus.append('*.*.bash_history')
        plus.append('*.*.profile')
        plus.append('*.DS_Store')
        return exs + plus

    def setExcludes(self, excludes=[]):
        self.excludes = excludes

    def makeFilters(self, folder, includes=[]):
        filters = []

        exs = folder.getExcludes() + self.getExcludes()
        for ex in exs:
            filters.append("- " + ex)

        inc = OrderedDict()
        for path in includes:
            p = ''
            parts = pathComponents(path)
            for part in parts:
                p = os.path.join(p, part)
                if p not in inc:
                    inc[p] = os.sep
                if part == parts[-1]:
                    inc[p] = os.sep + "***"

        for path, opt in inc.items():
            path = ''.join(path.rsplit(os.sep, 1)
                           ) if path.endswith(os.sep) else path
            filters.append("+ " + path + opt)

        filters.append("- *")
        return filters


class Rsync(object):

    def __init__(self):
        self.fromPath = None
        self.fromHost = None
        self.fromUsername = None
        self.toPath = None
        self.toHost = None
        self.toUsername = None

        self.filters = []
        self.opt = {
            'p': True,
            'a': True,
            'r': True,
            'u': True
        }
        self.longopt = {}

    def setTransport(self, port=22, keyfile=None):
        self.port = port
        self.keyfile = keyfile

    def setFrom(self, path, host=None, username=None):
        self.fromPath = path
        if not self.fromPath.endswith(os.sep):
            self.fromPath += os.sep
        self.fromHost = host
        self.fromUsername = username

    def setTo(self, path, host=None, username=None):
        self.toPath = path
        if not self.toPath.endswith(os.sep):
            self.toPath += os.sep
        self.toHost = host
        self.toUsername = username

    def setFilters(self, filters):
        self.filters = filters

    def setLongOption(self, option, v):
        if v:
            self.longopt[option] = v
        else:
            self.longopt.pop(option, None)

    def setToggleOption(self, option, v):
        if len(option) > 1:
            raise ValueError("Toggle options should be single letters.")

        if v:
            self.opt[option] = v
        else:
            self.opt.pop(option, None)

    def getCommandFilters(self):
        return "\n".join(self.filters)

    def getCommandOpt(self):
        return ''.join(list(self.opt.keys()))

    def getCommandLongOpt(self):
        opts = ""
        for k, v in self.longopt.items():
            if v is True:
                opts += " --%s" % k
            elif v:
                opts += " --%s=\"%s\"" % (k, v)
        return opts

    def getCommandFrom(self):
        dest = self.fromPath
        if self.fromHost:
            dest = self.fromHost + ":" + dest
            if self.fromUsername:
                dest = self.fromUsername + "@" + dest
        return dest

    def getCommandTo(self):
        dest = self.toPath
        if self.toHost:
            dest = self.toHost + ":" + dest
            if self.toUsername:
                dest = self.toUsername + "@" + dest
        return dest

    def getCommandTransport(self):
        transport = "ssh -p %s -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null" % self.port
        if self.keyfile is not None:
            transport += " -i \"%s\"" % self.keyfile
        return transport

    def getCommand(self):
        params = {
            'filters': self.getCommandFilters(),
            'opt': self.getCommandOpt(),
            'longopt': self.getCommandLongOpt(),
            'transport': self.getCommandTransport(),
            'from': self.getCommandFrom(),
            'to': self.getCommandTo()
        }
        return "echo '%(filters)s' | rsync -%(opt)se '%(transport)s' %(longopt)s --include-from=- %(from)s %(to)s" % params

    def sync(self):
        op = Shell.call(self.getCommand())
        if op.isFail():
            print(("%s" % op))
            return op
        return OK(self)



from setuptools import setup, find_packages
import platform

with open('README.rst') as f:
    readme = f.read()

with open('substance/_version.py') as versionFile:
    exec(versionFile.read())

install_requires = [
    'setuptools>=1.1.3',
    'PyYAML',
    'tabulate',
    'paramiko>=2.3.1',
    'netaddr',
    'requests',
    'tinydb',
    'python_hosts==0.3.3',
    'jinja2'
]

setup(name='substance',
      version=__version__,
      author='Turbulent',
      author_email='oss@turbulent.ca',
      url='https://substance.readthedocs.io/',
      license='Apache License 2.0',
      long_description=readme,
      description='Substance - Local dockerized development environment',
      install_requires=install_requires,
      extras_require={
        ':sys.platform == "darwin"': ['macfsevents'],
        ':sys.platform == "linux"': ['watchdog']
      },
      python_requires='>=3',
      packages=find_packages(),
      package_data={'substance': ['support/*']},
      test_suite='tests',
      zip_safe=False,
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'substance = substance.cli:cli',
              'subenv = substance.subenv.cli:cli'
          ],
      })



from substance.monads import *
from substance.logs import *
from substance import (Engine, Command)
from substance.exceptions import (SubstanceError)


class Ssh(Command):

    def getUsage(self):
        return "substance ssh [container?]"

    def getHelpTitle(self):
        return "Connect using SSH to the engine or directly into a container"

    def getShellOptions(self, optparser):
        optparser.add_option("-u", "--user", dest="user",
                             help="User to connect", default=None)
        optparser.add_option("-d", "--cwd", dest="cwd",
                             help="Change to this directory upon connect", default=None)
        return optparser

    def main(self):
        return self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.envShell, container=self.getInputContainer(), user=self.getOption('user'), cwd=self.getOption('cwd')) \
            .catch(self.exitError)

    def getInputContainer(self):
        if len(self.args) <= 0:
            return None
        return self.args[0]



from substance.monads import *
from substance.logs import *
from substance import (Command, Engine)
from tabulate import tabulate


class Switch(Command):
    def getShellOptions(self, optparser):
        optparser.add_option("-k", "--keep-containers", dest="keepcontainers",
                             help="Perform switch but do not restart containers", default=False, action="store_true")
        return optparser

    def getUsage(self):
        return "substance switch [options] SUBENV-NAME"

    def getHelpTitle(self):
        return "Switch the active subenv for an engine"

    def main(self):
        subenv = self.getArg(0)
        if not subenv:
            return self.exitError("Please specify a subenv name to switch to.")

        restart = not self.getOption('keepcontainers')
        self.core.loadCurrentEngine(name=self.parent.getOption('engine')) \
            .bind(Engine.loadConfigFile) \
            .bind(Engine.envSwitch, subenvName=subenv, restart=restart) \
            .catch(self.exitError)



from abc import (ABC, abstractmethod)
from substance.utils import (CommandList)


class Orchestrator(ABC):

    engine = None

    def __init__(self, engine):
        self.engine = engine
        super().__init__()

    def resetAll(self, time=10):
        return [
            "docker ps -q --filter 'label=ca.turbulent.dockwrkr.managed' | xargs -r docker rm -f",
            "docker ps -q --filter 'label=com.docker.compose.version' | xargs -r docker rm -f"
        ]

    @abstractmethod
    def startAll(self):
        pass

    @abstractmethod
    def restartAll(self, time=10):
        pass

    @abstractmethod
    def stopAll(self, time=10):
        pass

    @abstractmethod
    def removeAll(self, time=10):
        pass

    @abstractmethod
    def start(self, containers):
        pass

    @abstractmethod
    def stop(self, containers, time=10):
        pass

    @abstractmethod
    def remove(self, containers, time=10):
        pass

    @abstractmethod
    def restart(self, containers, time=10):
        pass

    @abstractmethod
    def exec(self, container, cmd, cwd=None, user=None):
        pass

    @abstractmethod
    def run(self, container, commands):
        pass

    @abstractmethod
    def status(self):
        pass


class Dockwrkr(Orchestrator):

    def login(self):
        return ["dockwrkr login"]

    def startAll(self):
        return ["dockwrkr start -a"]

    def restartAll(self, time=10):
        return ["dockwrkr restart -a -t %s" % (time)]

    def stopAll(self, time=10):
        return ["dockwrkr stop -a -t %s" % (time)]

    def removeAll(self, time=10):
        return ["dockwrkr remove -a -t %s" % (time)]

    def start(self, containers):
        return ["dockwrkr start %s" % (' '.join(containers))]

    def stop(self, containers, time=10):
        return ["dockwrkr stop %s -t %s" % (' '.join(containers), time)]

    def remove(self, container, time=10):
        return ["dockwrkr remove %s -t %s" % (' '.join(containers), time)]

    def restart(self, containers, time=10):
        return ["dockwrkr restart %s -t %s" % (' '.join(containers), time)]

    def recreate(self, containers, time=10):
        return ["dockwrkr recreate %s -t %s" % (' '.join(containers), time)]

    def exec(self, container, cmd, cwd=None, user=None):
        opts = []
        cmdline = ' '.join(cmd)
        initCommands = CommandList(['export TERM=xterm', cmdline])
        if user:
            opts.append('--user %s' % user)
        if cwd:
            initCommands.insert(0, 'cd %s' % cwd)

        tpl = {
            'opts': ' '.join(opts),
            'container': container,
            'initCommands': initCommands.logicAnd()
        }

        return ["dockwrkr exec -t -i %(opts)s %(container)s '/bin/sh -c \"%(initCommands)s\"'" % tpl]

    def run(self, task, args):
        return ["dockwrkr run %s %s" % (task, ' '.join(args))]

    def status(self):
        return ["dockwrkr status"]


class Compose(Orchestrator):

    def login(self):
        return []

    def startAll(self):
        return ["docker-compose up -d"]

    def restartAll(self, time=10):
        return ["docker-compose restart-t %s" % (time)]

    def stopAll(self, time=10):
        return ["docker-compose stop -t %s" % (time)]

    def removeAll(self, time=10):
        return ["docker-compose rm -f -s"]

    def start(self, containers):
        return ["dockwrkr start %s -d" % (' '.join(containers))]

    def stop(self, containers, time=10):
        return ["docker-compose stop %s -t %s" % (' '.join(containers), time)]

    def remove(self, container, time=10):
        return ["docker-compose rm -s -f %s" % (' '.join(containers))]

    def restart(self, containers, time=10):
        return ["docker-compose restart %s -t %s" % (' '.join(containers), time)]

    def recreate(self, containers, time=10):
        return ["(docker-compose rm -f -s %s && docker-compose start %s)" % (' '.join(containers), ' '.join(containers))]

    def exec(self, container, cmd, cwd=None, user=None):
        opts = []
        cmdline = ' '.join(cmd)
        initCommands = ['export TERM=xterm', cmdline]
        if user:
            opts.append('--user %s' % user)
        if cwd:
            initCommands.insert(0, 'cd %s' % cwd)

        tpl = {
            'opts': ' '.join(opts),
            'container': container,
            'initCommands': ' && '.join(initCommands)
        }

        return ["docker-compose exec %(opts)s %(container)s /bin/sh -c \"%(initCommands)s\"" % tpl]

    def run(self, task, args):
        return ["docker-compose run %s %s" % (task, ' '.join(args))]

    def status(self):
        return ["docker-compose ps"]



import logging
from substance.logs import DriverLogAdapter

logger = logging.getLogger(__name__)


class Driver(object):
    '''
    Substance virtualization driver abstract
    '''

    def __init__(self, core):
        self.core = core
        self.logAdapter = DriverLogAdapter(
            logger, {'name': self.__class__.__name__})

    def importMachine(self, name, ovfFile, engineProfile=None):
        '''Import a machine image into a virtual machine'''
        raise NotImplementedError()

    def deleteMachine(self, uuid):
        '''Delete a virtual machine by driver id'''
        raise NotImplementedError()

    def startMachine(self, uuid):
        '''Start a virtual machine by driver id'''
        raise NotImplementedError()

    def suspendMachine(self, uuid):
        '''Suspend a virtual machine by driver id'''
        raise NotImplementedError()

    def haltMachine(self, uuid):
        '''Safely halt a virtual machine by driver id'''
        raise NotImplementedError()

    def terminateMachine(self, uuid):
        '''Forcefully terminate a virtual machine by driver id'''
        raise NotImplementedError()

    def getMachines(self):
        '''
        Get a list of virtual machines available from this driver.
        Form is a [dict(name:uuid)]
        '''
        raise NotImplementedError()

    def getMachineID(self, name):
        '''Retrieve the driver identifier for a machine name'''
        raise NotImplementedError()

    def getMachineState(self, uuid):
        '''Get the substance engine state for a virtual machine by driver id'''
        raise NotImplementedError()

    def exists(self, uuid):
        '''Check for existence of a virtual machine by driver id'''
        raise NotImplementedError()

    def isRunning(self, uuid):
        '''Check for that a virtual machine is running by driver id'''
        raise NotImplementedError()

    def isStopped(self, uuid):
        '''Check for that a virtual machine is stopped by driver id'''
        raise NotImplementedError()

    def isSuspended(self, uuid):
        '''Check for that a virtual machine is suspended by driver id'''
        raise NotImplementedError()



import os
import logging
import requests
from collections import OrderedDict
from substance.monads import *
from substance.logs import *
from substance.exceptions import *
from substance.shell import Shell
from substance.utils import (
    readYAML,
    writeYAML,
    readSupportFile,
    getSupportFile,
    streamDownload,
    makeXHRRequest,
    sha1sum,
    untar
)

logger = logging.getLogger(__name__)


class Box(object):

    def __init__(self, core, name, version, namespace, registry=None, boxstring=None, archiveSHA1=None):
        self.core = core
        self.name = name
        self.version = version
        self.namespace = namespace
        self.registry = registry
        self.boxstring = boxstring
        self.archiveSHA1 = archiveSHA1

    def getRegistryURL(self):
        return "https://%(registry)s/%(namespace)s/%(name)s/%(version)s.json" % self.__dict__

    def getArchiveURL(self):
        return "https://%(registry)s/%(namespace)s/%(name)s/%(version)s.box" % self.__dict__

    def getDescriptor(self):
        return {'name': self.name, 'registry': self.registry, 'namespace': self.namespace}

    def getPath(self):
        parts = [self.core.getBoxesPath(), self.registry, self.namespace,
                 self.name, self.version]
        return os.path.join(*parts)

    def getFilePath(self, filename):
        return os.path.join(self.getPath(), filename)

    def getArchivePath(self):
        return self.getFilePath("box.box")

    def getOVFFile(self):
        return self.getFilePath("box.ovf")

    def getImageName(self):
        return "%(name)s:%(version)s" % self.__dict__

    @staticmethod
    def fromBoxString(name):
        return Box.parseBoxString(name).map(lambda b: Box(**b))

    @staticmethod
    def parseBoxString(name):
        '''
        Parse a box string into dict of the components.

        Format: namespace/name:VERSION@registry
        registry and version are optional.
        '''

        box = OrderedDict()

        box['boxstring'] = name
        parts = name.split("@", 1)
        boxstring = parts[0]

        if len(parts) > 1:
            box['registry'] = parts[1]
        else:
            box['registry'] = "dl.bintray.com/turbulent/substance-images"

        parts = boxstring.split("/", 1)
        if len(parts) <= 1:
            return Fail(InvalidBoxName("Name \"%s\" does not reference a valid box name: Missing namespace." % name))

        (box['namespace'], imagestring) = parts
        # substance-box:0.1

        imageparts = imagestring.split(":", 1)
        if len(imageparts) > 1:
            (box['name'], box['version']) = imageparts
        else:
            box['name'] = imageparts[0]
            box['version'] = 'latest'

        return OK(box)

    def fetch(self):
        if self.isDownloaded():
            return OK(self)

        manifestURL = self.getRegistryURL()
        logger.debug("Downloading manifest: %s", manifestURL)
        return Try.attempt(makeXHRRequest, url=manifestURL) \
            .catch(lambda err: Fail(InvalidBoxName("Failed to fetch the box manifest for %s. Does this box exist?" % self.boxstring))) \
            .bind(defer(self.download))

    def isDownloaded(self):
        return True if os.path.exists(self.getOVFFile()) else False

    def getShortBoxString(self):
        return "%s/%s:%s" % (self.namespace, self.name, self.version)

    def delete(self):
        logger.info("Removing box %s from box cache", self.getShortBoxString())
        return Try.sequence([
            self.core.getDB().removeBoxRecord(self),
            Shell.nukeDirectory(self.getPath())
        ])

    def download(self, boxResult):
        archiveURL = boxResult['archiveURL']
        archiveSHA = boxResult['archiveSHA1']
        archive = self.getArchivePath()

        logger.info("Downloading %s:%s (%s)", self.name,
                    self.version, boxResult['archiveURL'])

        return Try.sequence([
            Shell.makeDirectory(os.path.dirname(archive)),
            Try.attempt(streamDownload, url=archiveURL, destination=archive)
            .then(defer(self.verifyArchive, expectedSHA=archiveSHA))
            .then(defer(self.unpackArchive)),
            self.core.getDB().updateBoxRecord(self, boxResult)
        ]).map(lambda x: self)

    def getArchiveHash(self):
        return sha1sum(self.getArchivePath())

    def verifyArchive(self, expectedSHA):
        logger.info("Verifying archive for %s", self.getImageName())
        sha = self.getArchiveHash()
        if sha == expectedSHA:
            return OK(self)
        else:
            Try.attempt(lambda: os.remove(self.getArchivePath()))
            return Fail(BoxArchiveHashMismatch("Box archive hash mismatch. Failed download?"))

    def unpackArchive(self):
        return Try.sequence([
            Try.attempt(untar, self.getArchivePath(), self.getPath()),
            Shell.rmFile(self.getArchivePath())
        ])




import sys

# from .constants import *
# from .exceptions import *
# from .monads import *
# from .logs import *

from .shell import Shell

from .core import Core
from .db import DB
from .box import Box

from .engine import Engine, EngineProfile
from .driver import Driver
from .link import Link

from .command.command import Command
from .command.command import Program
from .command.command import SubProgram
